# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the paginator extensions."""

from typing import override

from django.core.exceptions import ImproperlyConfigured
from django.core.paginator import PageNotAnInteger
from django.template import engines
from django.test import override_settings

from debusine.db.models import User
from debusine.web.views.table import Column, FilterText, Table
from debusine.web.views.table.tests.test_table import TableTestCase


class PaginatorTests(TableTestCase):
    """Tests for :py:class:`Paginator`."""

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        for idx in range(10):
            cls.playground.create_user(f"user{idx:02d}")

    @override
    def get_table_class(self) -> type[Table[User]]:
        class _Table(Table[User]):
            user = Column("User", ordering="username")
            default_order = "user"

        return _Table

    def test_bool_has_rows(self) -> None:
        """Test __bool__."""
        p = self._table().get_paginator(per_page=3)
        self.assertTrue(p)

    def test_bool_has_no_rows(self) -> None:
        """Test __bool__."""
        User.objects.all().delete()
        p = self._table().get_paginator(per_page=3)
        self.assertFalse(p)

    def test_pages(self) -> None:
        """Test with no columns defined."""
        p = self._table().get_paginator(per_page=3)
        self.assertEqual(p.num_pages, 4)
        self.assertFalse(p.page_obj.has_previous())

    def test_has_page_navigation(self) -> None:
        """Test has_page_navigation attribute."""
        table = self._table()
        for page_size, expected in (
            (3, True),
            (10, True),
            (11, False),
            (12, False),
        ):
            with self.subTest(page_size=page_size):
                p = table.get_paginator(per_page=page_size)
                self.assertEqual(p.has_page_navigation, expected)

    def test_render(self) -> None:
        """Test rendering as a widget."""
        with self.template_dir() as template_dir:
            (template_dir / "test.html").write_text("{{paginator.per_page}}")
            table = self._table()
            table.template_name = "test.html"
            p = table.get_paginator(per_page=42)
            rendered = self.render_widget(p)
            self.assertEqual(rendered, "42")

    def test_render_thead(self) -> None:
        """Test rendering <thead> as a widget."""
        with self.template_dir() as template_dir:
            (template_dir / "test.html").write_text(
                "{% load debusine %}{% widget paginator.thead %}"
            )
            table = self._table()
            table.template_name = "test.html"
            p = table.get_paginator(per_page=42)
            tree = self.assertHTMLValid(self.render_widget(p))
            thead = self.assertHasElement(tree, "body/thead")
            # There are no filters, so there is only one row
            self.assertEqual(len(thead.tr), 1)
            self.assertEqual(
                [
                    self.get_node_text_normalized(th.div.div[0])
                    for th in thead.tr.th
                ],
                ["User"],
            )

    def test_render_thead_filters(self) -> None:
        """Test rendering <thead> as a widget, with filters."""

        class _Table(Table[User]):
            user = Column("User", ordering="username")
            default_order = "user"

            filter_username = FilterText("User")

        with self.template_dir() as template_dir:
            (template_dir / "test.html").write_text(
                "{% load debusine %}{% widget paginator.thead %}"
            )
            table = self._table(table_class=_Table)
            table.template_name = "test.html"
            p = table.get_paginator(per_page=42)
            tree = self.assertHTMLValid(self.render_widget(p))
            thead = self.assertHasElement(tree, "body/thead")
            self.assertEqual(len(thead.tr), 2)
            self.assertEqual(thead.tr[0].th.get("colspan"), "1")
            self.assertEqual(
                [
                    self.get_node_text_normalized(th.div.div[0])
                    for th in thead.tr[1].th
                ],
                ["User"],
            )

    def test_render_thead_preview_mode(self) -> None:
        """Test rendering <thead> as a widget."""

        class _Table(Table[User]):
            user = Column("User", ordering="username")
            default_order = "user"

            filter_username = FilterText("User")

        with self.template_dir() as template_dir:
            (template_dir / "test.html").write_text(
                "{% load debusine %}{% widget paginator.thead %}"
            )
            table = self._table(table_class=_Table, preview=True)
            table.template_name = "test.html"
            p = table.get_paginator(per_page=3)
            tree = self.assertHTMLValid(self.render_widget(p))
            thead = self.assertHasElement(tree, "body/thead")
            # In preview mode, the filters are hidden
            self.assertEqual(len(thead.tr), 1)
            self.assertEqual(
                [
                    self.get_node_text_normalized(th.div.div[0])
                    for th in thead.tr.th
                ],
                ["User"],
            )

    def test_render_tfoot(self) -> None:
        """Test rendering <tfoot> as a widget."""
        with self.template_dir() as template_dir:
            (template_dir / "test.html").write_text(
                "{% load debusine %}{% widget paginator.tfoot %}"
            )
            table = self._table()
            table.template_name = "test.html"
            p = table.get_paginator(per_page=10)
            tree = self.assertHTMLValid(self.render_widget(p))
            tfoot = self.assertHasElement(tree, "body/tfoot")
            tr = self.assertHasElement(tfoot, "tr")
            self.assertEqual(tr.td.get("colspan"), "1")

    def test_tfoot_preview_mode(self) -> None:
        """Table footer in preview mode shows the preview footer."""
        with self.template_dir() as template_dir:
            (template_dir / "test.html").write_text(
                "{% load debusine %}{% widget paginator.tfoot %}"
            )
            table = self._table(preview=True)
            table.template_name = "test.html"
            p = table.get_paginator(per_page=3)
            tree = self.assertHTMLValid(self.render_widget(p))
            tfoot = self.assertHasElement(tree, "body/tfoot")
            tr = self.assertHasElement(tfoot, "tr")
            self.assertEqual(tr.td.get("colspan"), "1")
            self.assertTextContentEqual(tr.td, "3 out of 11 shown")

    def test_tfoot_preview_mode_no_pagination(self) -> None:
        """Table footer in preview mode footer not shown if one page only."""
        with self.template_dir() as template_dir:
            (template_dir / "test.html").write_text(
                "{% load debusine %}{% widget paginator.tfoot %}"
            )
            table = self._table(preview=True)
            table.template_name = "test.html"
            p = table.get_paginator(per_page=50)
            self.assertEqual(self.render_widget(p).strip(), "")

    def test_render_without_template_name(self) -> None:
        """Test rendering as a widget."""
        table = self._table()
        p = table.get_paginator(per_page=42)

        template = engines["django"].from_string(
            "{% load debusine %}{% widget widget %}"
        )
        setattr(template, "engine", engines["django"])
        # Set DEBUG=True to have render raise exceptions instead of logging
        # them
        with (
            self.assertRaisesRegex(
                ImproperlyConfigured,
                "Table.template_name needs to be set"
                " to use Paginator as a widget",
            ),
            override_settings(DEBUG=True),
        ):
            template.render({"widget": p})

    def test_validate_number(self) -> None:
        """Test clamping behaviour of validate_number override."""
        table = self._table()
        paginator = table.get_paginator(per_page=50)
        num_pages = paginator.num_pages

        with self.subTest(case="valid_page"):
            self.assertEqual(paginator.validate_number(num_pages), num_pages)

        with self.subTest(case="overflow_clamped_to_last"):
            self.assertEqual(
                paginator.validate_number(num_pages + 1), num_pages
            )

        with (
            self.subTest(case="float_input_raises"),
            self.assertRaises(PageNotAnInteger),
        ):
            paginator.validate_number(num_pages + 1.5)

        with (
            self.subTest(case="non_numeric_string_raises"),
            self.assertRaises(PageNotAnInteger),
        ):
            paginator.validate_number("Invalid page number")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""GnuPG utilities for the Debusine signing service."""

import os
import shutil
import subprocess
from collections.abc import Generator, Iterable
from contextlib import ExitStack, contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import BinaryIO, TYPE_CHECKING

import gpg

from debusine.signing.models import (
    AvailableKeyFileSystem,
    ProtectedKey,
    SigningMode,
)
from debusine.signing.signing_utils import SensitiveTemporaryDirectory

if TYPE_CHECKING:
    from _typeshed import ReadableBuffer

    ReadableBuffer  # fake usage for vulture


class GnuPGError(Exception):
    """A GnuPG operation failed."""


# Based loosely on EphemeralContext in gpgme's lang/python/tests/support.py.
@contextmanager
def gpg_ephemeral_context(
    temp_dir: Path,
) -> Generator[gpg.Context]:
    """Run gpg, cleaning up properly afterwards."""
    home = temp_dir / "gpg"
    home.mkdir(mode=0o700)
    try:
        (home / "gpg.conf").write_text(
            # Prefer a SHA-2 hash that's long enough to use with Ed25519.
            # This roughly matches GnuPG's defaults, except without SHA-1
            # and SHA-224; and it seems that Debian's patch to prefer
            # SHA-512 when signing doesn't apply to all key types.
            "personal-digest-preferences SHA512 SHA384 SHA256\n"
        )
        with gpg.Context(home_dir=str(home)) as ctx:
            ctx.armor = True
            yield ctx
            gpgconf_env = os.environ.copy()
            gpgconf_env["GNUPGHOME"] = str(home)
            subprocess.run(["gpgconf", "--kill", "all"], env=gpgconf_env)
    finally:
        shutil.rmtree(home, ignore_errors=True)


def gpg_generate(description: str) -> tuple[bytes, bytes]:
    """Generate a new key pair using GnuPG."""
    with (
        SensitiveTemporaryDirectory("debusine-gnupg-generate-") as tmp,
        gpg_ephemeral_context(Path(tmp)) as ctx,
    ):
        # TODO: Ed25519 is good enough for most purposes, but we should
        # really allow the caller to set the algorithm and key length.  In
        # particular, GnuPG 1 doesn't support Ed25519, so if we need
        # compatibility with that then we'd need to support something like
        # RSA.
        key = ctx.create_key(
            description, algorithm="ed25519", expires=False, sign=True
        )
        if not key.primary:
            raise GnuPGError("Failed to generate a secret key")
        if key.sub:
            raise GnuPGError("Unexpectedly got an encryption subkey")
        secret = ctx.key_export_secret(key.fpr)
        if not secret:
            raise GnuPGError("Failed to export secret key")
        public = ctx.key_export_minimal(key.fpr)
        if not public:
            raise GnuPGError("Failed to export public key")
        return secret, public


def gpg_fingerprint(public_key: "ReadableBuffer") -> str:
    """Get the fingerprint of a GnuPG public key."""
    with (
        TemporaryDirectory("debusine-gnupg-import-") as tmp,
        gpg_ephemeral_context(Path(tmp)) as ctx,
    ):
        result = ctx.key_import(public_key)
        if not getattr(result, "imported", 0):
            raise GnuPGError(f"Failed to get fingerprint of new key: {result}")
        assert isinstance(result.imports[0].fpr, str)
        return result.imports[0].fpr


def gpg_sign(
    keys: Iterable[tuple[ProtectedKey, "ReadableBuffer"]],
    data_path: Path,
    signature_path: Path,
    mode: SigningMode = SigningMode.CLEAR,
) -> None:
    """Sign data using GnuPG."""
    match mode:
        case SigningMode.DETACHED:
            gpg_mode = gpg.constants.SIG_MODE_DETACH
        case SigningMode.CLEAR:
            gpg_mode = gpg.constants.SIG_MODE_CLEAR
        case _:
            raise GnuPGError(
                "OpenPGP signing currently only supports detached or clear "
                "signatures"
            )

    with ExitStack() as stack:
        availables: list[AvailableKeyFileSystem] = []
        for private_key, public_key in keys:
            available = stack.enter_context(private_key.available(public_key))
            if not isinstance(available, AvailableKeyFileSystem):
                raise GnuPGError(
                    "OpenPGP signing currently only supports "
                    "software-encrypted keys"
                )
            availables.append(available)

        with (
            SensitiveTemporaryDirectory("debusine-gnupg-sign-") as tmp,
            gpg_ephemeral_context(Path(tmp)) as ctx,
        ):
            signers = []
            for available in availables:
                import_result = ctx.key_import(available._key_path.read_bytes())
                if not getattr(import_result, "secret_imported", 0):
                    raise GnuPGError("Failed to import secret key")
                signers += ctx.keylist(
                    pattern=import_result.imports[0].fpr, secret=True
                )
            ctx.signers = signers

            with (
                open(data_path, mode="rb") as data_file,
                open(signature_path, mode="wb") as signature_file,
            ):
                ctx.sign(
                    gpg.Data(file=data_file), sink=signature_file, mode=gpg_mode
                )


def gpg_debsign(
    private_key: ProtectedKey,
    public_key: "ReadableBuffer",
    fingerprint: str,
    data_path: Path,
    signature_path: Path,
    log_file: BinaryIO | None = None,
) -> None:
    """
    Sign an upload using GnuPG and `debsign`.

    :param data_path: path to the unsigned `.changes` file.  Other files
      referenced by it that need to be signed must be in the same directory.
    :param signature_path: path to the signed `.changes` file.  Other files
      referenced by it that were signed are in the same directory.
    """
    with private_key.available(public_key) as available:
        if not isinstance(available, AvailableKeyFileSystem):
            raise GnuPGError(
                "OpenPGP signing currently only supports software-encrypted "
                "keys"
            )
        with (
            SensitiveTemporaryDirectory("debusine-debsign-") as tmp,
            gpg_ephemeral_context(Path(tmp)) as ctx,
        ):
            import_result = ctx.key_import(available._key_path.read_bytes())
            if not getattr(import_result, "secret_imported", 0):
                raise GnuPGError("Failed to import secret key")

            if list(signature_path.parent.iterdir()):
                raise GnuPGError(
                    "Directory containing signature_path is not empty"
                )
            shutil.copytree(
                data_path.parent, signature_path.parent, dirs_exist_ok=True
            )

            subprocess.run(
                ["debsign", f"-k{fingerprint}", "--re-sign", signature_path],
                cwd=signature_path.parent,
                env={**os.environ, "GNUPGHOME": str(ctx.home_dir)},
                check=True,
                stdout=log_file,
                stderr=log_file,
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for auth-related models."""

import contextlib
import datetime as dt
import re
from collections.abc import Generator
from typing import Any, ClassVar, override
from unittest import mock

from asgiref.sync import sync_to_async
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured
from django.db import IntegrityError, ProgrammingError, transaction
from django.test import RequestFactory
from django.utils import timezone

import debusine.db.models.auth as auth
from debusine.db.context import ContextConsistencyError, context
from debusine.db.models import (
    Asset,
    Group,
    Identity,
    Scope,
    Token,
    User,
    Worker,
    Workspace,
)
from debusine.db.models.auth import (
    ClientEnroll,
    GroupAuditLog,
    GroupAuditLogEntryTypes,
    GroupMembership,
    GroupRoleBase,
    GroupRoles,
    group_audit_log_entry_adapter,
    system_user,
)
from debusine.db.models.tests.utils import RolesTestCase
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.test.django import (
    ChannelsHelpersMixin,
    TestCase,
    TransactionTestCase,
)


class TokenTests(ChannelsHelpersMixin, TestCase):
    """Unit tests of the ``Token`` model."""

    scenario = scenarios.DefaultScopeUser()

    def test_save(self) -> None:
        """The model creates a Token.key on save or keeps it if it existed."""
        token = self.playground.create_user_token()

        self.assertIsNotNone(token.id)
        self.assertEqual(len(token.key), 64)

        key = token.key
        token.save()
        self.assertEqual(token.key, key)

    def test_str(self) -> None:
        """Test Token.__str__."""
        token = self.playground.create_user_token()
        self.assertEqual(token.__str__(), token.hash)

    def test_external_user_inconsistent_grants(self) -> None:
        with self.assertRaisesRegex(
            IntegrityError, "db_token_consistent_grants"
        ):
            Token.objects.create(token_type=Token.Types.EXTERNAL_USER)

    def test_get_token_or_none_found(self) -> None:
        """get_token_or_none looks up a token and returns it."""
        token_hash = Token._generate_hash('some_key')
        token = self.playground.create_user_token(hash=token_hash)

        got_token = Token.objects.get_token_or_none("some_key")
        assert got_token is not None
        self.assertEqual(got_token, token)

        # The worker is prefetched.  This is important because hasattr looks
        # like a harmless thing to call from async code.
        with self.assertNumQueries(0):
            self.assertFalse(hasattr(got_token, "worker"))
        with self.assertNumQueries(0):
            self.assertFalse(hasattr(got_token, "activating_worker"))
        with self.assertNumQueries(0):
            self.assertIsNone(got_token.work_request)

    def test_get_token_or_none_expires_soon(self) -> None:
        """get_token_or_none allows tokens with `expire_at` in the future."""
        token_hash = Token._generate_hash('some_key')
        token = self.playground.create_user_token(
            hash=token_hash, expire_at=timezone.now() + dt.timedelta(days=1)
        )

        self.assertEqual(token, Token.objects.get_token_or_none('some_key'))

    def test_get_token_or_none_expired(self) -> None:
        """get_token_or_none skips tokens with `expire_at` in the past."""
        token_hash = Token._generate_hash('some_key')
        self.playground.create_user_token(
            hash=token_hash, expire_at=timezone.now() - dt.timedelta(seconds=1)
        )

        self.assertIsNone(Token.objects.get_token_or_none('some_key'))

    def test_get_token_or_none_not_found(self) -> None:
        """get_token_or_none cannot find a token and returns None."""
        self.assertIsNone(Token.objects.get_token_or_none('a_non_existing_key'))

    def test_get_token_worker_prefetch(self) -> None:
        """get_token_or_none prefetches the worker attribute if present."""
        token_hash = Token._generate_hash('some_key')
        token = self.playground.create_worker_token(
            set_worker=False, hash=token_hash
        )
        worker = self.playground.create_worker(token=token)

        got_token = Token.objects.get_token_or_none("some_key")
        self.assertEqual(got_token, token)
        assert got_token is not None

        # worker is prefetched
        with self.assertNumQueries(0):
            self.assertEqual(got_token.worker, worker)

    def test_get_token_activating_worker_prefetch(self) -> None:
        """get_token_or_none prefetches activating_worker if present."""
        token_hash = Token._generate_hash('some_key')
        token = self.playground.create_worker_activation_token(
            set_worker=False, hash=token_hash
        )
        activating_worker = self.playground.create_worker(
            activation_token=token
        )

        got_token = Token.objects.get_token_or_none("some_key")
        self.assertEqual(got_token, token)
        assert got_token is not None

        # worker is prefetched
        with self.assertNumQueries(0):
            self.assertEqual(got_token.activating_worker, activating_worker)

    def test_get_token_work_request__worker_prefetch(self) -> None:
        """get_token_or_none prefetches workrequest_worker.worker."""
        work_request = self.playground.create_worker_task(
            assign_new_worker=True, mark_running=True
        )
        token = self.playground.create_work_request_token(work_request)

        got_token = Token.objects.get_token_or_none(token.key)
        self.assertEqual(got_token, token)
        assert got_token is not None

        # worker is prefetched
        with self.assertNumQueries(0):
            assert got_token.work_request is not None
            self.assertEqual(got_token.work_request.worker, work_request.worker)

    def test_get_token_work_request_token__work_request__created_by_prefetch(
        self,
    ) -> None:
        """get_token_or_none prefetches all the way through to created_by."""
        work_request = self.playground.create_worker_task(
            assign_new_worker=True, mark_running=True
        )
        token = self.playground.create_work_request_token(work_request)

        got_token = Token.objects.get_token_or_none(token.key)
        self.assertEqual(token, got_token)
        assert got_token is not None

        # worker is prefetched
        with self.assertNumQueries(0):
            assert got_token.work_request is not None
            self.assertEqual(
                got_token.work_request.created_by,
                work_request.created_by,
            )

    def test_enable(self) -> None:
        """enable() enables the token."""
        token = Token.objects.create(
            token_type=Token.Types.USER, user=self.scenario.user
        )

        # Assert the default is disabled tokens
        self.assertFalse(token.enabled)

        token.enable()
        token.refresh_from_db()

        self.assertTrue(token.enabled)

    async def test_disable(self) -> None:
        """disable() disables the token."""
        token = await Token.objects.acreate(
            token_type=Token.Types.WORKER, enabled=True
        )
        await Worker.objects.acreate(token=token, registered_at=timezone.now())
        await token.arefresh_from_db()

        channel = await self.create_channel(token.hash)

        await sync_to_async(token.disable)()

        await self.assert_channel_received(channel, {"type": "worker.disabled"})
        await token.arefresh_from_db()

        self.assertFalse(token.enabled)

    def test_get_username(self) -> None:
        for token, expected_username in (
            (self.playground.create_worker_token(), ""),
            (self.playground.create_worker_activation_token(), ""),
            (self.playground.create_work_request_token(), ""),
            (self.playground.create_user_token(), self.scenario.user.username),
            (
                self.playground.create_external_user_token(username="test"),
                "test",
            ),
        ):
            with self.subTest(token_type=str(token.token_type)):
                self.assertEqual(token.get_username(), expected_username)


class TokenTransactionTests(TransactionTestCase):
    """Unit tests for Token that require transaction finalization."""

    def test_foreign_key_constraints(self) -> None:
        for token_type in Token.Types:
            kwargs: dict[str, Any] = {}
            if token_type == Token.Types.EXTERNAL_USER:
                kwargs["explicit_grants_only"] = True
            try:
                with (
                    self.subTest(token_type=str(token_type)),
                    self.assertRaises(ProgrammingError),
                    transaction.atomic(),
                ):
                    token = Token.objects.create(
                        token_type=token_type, **kwargs
                    )
            except AssertionError:  # pragma: no cover
                # In case the test failed and we created a token. This is a
                # noop if it was not committed.
                token.delete()


class TokenManagerTests(TestCase):
    """Unit tests for the ``TokenManager`` class."""

    user_john: ClassVar[User]
    user_bev: ClassVar[User]
    token_john: ClassVar[Token]
    token_bev: ClassVar[Token]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user_john = get_user_model().objects.create_user(
            username="John", email="john@example.com"
        )
        cls.user_bev = get_user_model().objects.create_user(
            username="Bev", email="bev@example.com"
        )
        cls.token_john = cls.playground.create_user_token(user=cls.user_john)
        cls.token_bev = cls.playground.create_user_token(user=cls.user_bev)

    def test_create_worker_activation(self) -> None:
        """create_worker_activation creates a short-lived token."""
        with self.freeze_time() as now, transaction.atomic():
            token = Token.objects.create_worker_activation()
            self.playground.create_worker(activation_token=token)
        self.assertEqual(token.expire_at, now + dt.timedelta(minutes=15))
        self.assertIsInstance(token.key, str)
        self.assertTrue(token.enabled)

    def test_create_work_request_token(self) -> None:
        work_request = self.playground.create_worker_task(
            mark_pending=True, assign_new_worker=True
        )

        with self.freeze_time() as now:
            token = Token.objects.create_work_request_token(work_request)

        self.assertEqual(token.expire_at, now + dt.timedelta(days=7))
        self.assertEqual(token.token_type, Token.Types.WORK_REQUEST)
        self.assertIsInstance(token.key, str)
        self.assertTrue(token.enabled)
        token.refresh_from_db()  # Fire the triggers
        self.assertEqual(token.work_request, work_request)

    def test_create_work_request_token_unassigned(self) -> None:
        work_request = self.playground.create_worker_task(mark_pending=True)
        with self.assertRaisesRegex(
            AssertionError,
            (
                r"Cannot issue a token for a WorkRequest that is not assigned "
                r"to a worker\."
            ),
        ):
            Token.objects.create_work_request_token(work_request)

    def test_create_work_request_token_replaces_previous_work_request_token(
        self,
    ) -> None:
        work_request = self.playground.create_worker_task(
            mark_pending=True, assign_new_worker=True
        )
        token1 = Token.objects.create_work_request_token(work_request)
        Token.objects.create_work_request_token(work_request)

        with self.assertRaises(Token.DoesNotExist):
            token1.refresh_from_db()

    def test_create_work_request_token_replaces_previous_token_for_worker(
        self,
    ) -> None:
        """Only one WorkRequestToken can exist at a time for each worker."""
        worker = self.playground.create_worker()
        work_request1 = self.playground.create_worker_task(
            mark_pending=True, worker=worker
        )
        token1 = Token.objects.create_work_request_token(work_request1)

        work_request2 = self.playground.create_worker_task(
            mark_pending=True, worker=worker
        )
        Token.objects.create_work_request_token(work_request2)

        with self.assertRaises(Token.DoesNotExist):
            token1.refresh_from_db()

    def test_create_work_request_token_blocked(self) -> None:
        work_request = self.playground.create_worker_task()
        with self.assertRaisesRegex(
            AssertionError,
            r"Cannot issue a token for a WorkRequest that has status blocked",
        ):
            Token.objects.create_work_request_token(work_request)

    def test_create_external_user_token(self) -> None:
        token = Token.objects.create_external_user_token(
            username="external-user",
            created_by=self.playground.get_default_user(),
            comment="test comment",
        )

        self.assertEqual(token.token_type, Token.Types.EXTERNAL_USER)
        self.assertEqual(token.external_username, "external-user")
        self.assertTrue(token.explicit_grants_only)
        self.assertEqual(token.comment, "test comment")

    def test_get_tokens_all(self) -> None:
        """get_tokens returns all the tokens if no filter is applied."""
        self.assertQuerySetEqual(
            Token.objects.get_tokens(),
            {self.token_bev, self.token_john},
            ordered=False,
        )

    def test_get_tokens_by_owner(self) -> None:
        """get_tokens returns the correct tokens when filtering by owner."""
        self.assertQuerySetEqual(
            Token.objects.get_tokens(username='John'), [self.token_john]
        )
        self.assertQuerySetEqual(
            Token.objects.get_tokens(username='Bev'), [self.token_bev]
        )
        self.assertQuerySetEqual(
            Token.objects.get_tokens(username='Someone'), []
        )

    def test_get_tokens_by_key(self) -> None:
        """get_tokens returns the correct tokens when filtering by key."""
        self.assertQuerySetEqual(
            Token.objects.get_tokens(key=self.token_john.key),
            [self.token_john],
        )
        self.assertQuerySetEqual(
            Token.objects.get_tokens(key='non-existing-key'), []
        )

    def test_get_tokens_by_key_owner_empty(self) -> None:
        """
        get_tokens returns nothing if using a key and username without matches.

        Key for the key parameter or username for the user parameter exist
        but are for different tokens.
        """
        self.assertQuerySetEqual(
            Token.objects.get_tokens(
                key=self.token_john.key, username=self.user_bev.username
            ),
            [],
        )

    def test_get_tokens_include_expired(self) -> None:
        """get_tokens only includes expired tokens if requested."""
        self.token_john.expire_at = timezone.now() - dt.timedelta(seconds=1)
        self.token_john.save()
        self.assertQuerySetEqual(Token.objects.get_tokens(username="John"), [])
        self.assertQuerySetEqual(
            Token.objects.get_tokens(username="John", include_expired=True),
            [self.token_john],
        )

    def test_expired(self) -> None:
        """``expired`` returns only the expired tokens."""
        now = timezone.now()
        tokens = [
            self.playground.create_user_token(expire_at=expire_at)
            for expire_at in (
                None,
                now - dt.timedelta(minutes=1),
                now - dt.timedelta(seconds=1),
                now,
                now + dt.timedelta(seconds=1),
            )
        ]

        self.assertQuerySetEqual(
            Token.objects.expired(at=now), tokens[1:3], ordered=False
        )
        self.assertQuerySetEqual(
            Token.objects.expired(at=now + dt.timedelta(seconds=1)),
            tokens[1:4],
            ordered=False,
        )


class UserTests(TestCase):
    """Tests for the User class."""

    scenario = scenarios.DefaultScopeUser()
    user: ClassVar[User]
    system_user: ClassVar[User]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.user = cls.playground.create_user("test")
        cls.system_user = system_user()

    def test_can_display_anonymous(self) -> None:
        """Test can_display with anonymous user."""
        self.assertPermission(
            "can_display",
            AnonymousUser(),
            denied=[self.scenario.user, self.user, self.system_user],
        )

    def test_can_display(self) -> None:
        """Test the can_display predicate."""
        self.assertPermission(
            "can_display",
            [self.scenario.user, self.user, self.system_user],
            allowed=[self.scenario.user, self.user, self.system_user],
        )

    def test_can_manage_anonymous(self) -> None:
        """Test can_manage with anonymous user."""
        self.assertPermission(
            "can_manage",
            AnonymousUser(),
            denied=[self.scenario.user, self.user, self.system_user],
        )

    def test_can_manage(self) -> None:
        """Test can_manage."""
        self.assertPermission(
            "can_manage",
            self.scenario.user,
            allowed=self.scenario.user,
            denied=[self.system_user, self.user],
        )
        self.assertPermission(
            "can_manage",
            self.user,
            allowed=self.user,
            denied=[self.system_user, self.scenario.user],
        )

    def test_ui(self) -> None:
        from debusine.web.views.ui.auth import UserUI

        request = RequestFactory().get("/")
        ui = self.scenario.user.ui(request)
        self.assertIsInstance(ui, UserUI)

    def test_get_absolute_url(self) -> None:
        """Test the get_absolute_url method."""
        self.assertEqual(
            self.scenario.user.get_absolute_url(),
            f"/-/user/{self.scenario.user.username}/",
        )

    def test_disallow_numeric_usernames(self) -> None:
        with self.assertRaisesRegex(
            IntegrityError, r"db_user_non_integer_username"
        ):
            self.playground.create_user("123")


class IdentityTests(TestCase):
    """Tests for the Identity class."""

    def test_str(self) -> None:
        """Stringification should show the unique key."""
        ident = Identity(issuer="salsa", subject="test@debian.org")
        self.assertEqual(str(ident), "salsa:test@debian.org")


class GroupRolesTests(RolesTestCase[GroupRoles]):
    """Tests for the GroupRoles class."""

    scenario = scenarios.DefaultContext()
    roles_class = GroupRoles

    def test_invariants(self) -> None:
        self.assertRolesInvariants()

    def test_choices(self) -> None:
        self.assertChoices()

    def test_init(self) -> None:
        r = GroupRoleBase("test")
        r._setup()
        self.assertEqual(r.implied_by_scope_roles, frozenset())
        self.assertEqual(r.implied_by_group_roles, frozenset([r]))

    def test_init_invalid_implication(self) -> None:
        r = GroupRoleBase("test", implied_by=[Asset.Roles.OWNER])
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"Group roles do not support implications"
            r" by <enum 'AssetRoles'>",
        ):
            r._setup()

    def test_values(self) -> None:
        self.assertEqual(Group.Roles.ADMIN.label, "Admin")
        self.assertEqual(
            Group.Roles.ADMIN.description,
            "Add and remove other users",
        )
        self.assertEqual(
            Group.Roles.ADMIN.implied_by_scope_roles,
            frozenset([Scope.Roles.OWNER]),
        )
        self.assertEqual(
            Group.Roles.ADMIN.implied_by_group_roles,
            frozenset([Group.Roles.ADMIN]),
        )

        self.assertEqual(Group.Roles.MEMBER.label, "Member")
        self.assertEqual(
            Group.Roles.MEMBER.description,
            "Member",
        )
        self.assertEqual(
            Group.Roles.MEMBER.implied_by_scope_roles,
            frozenset([Scope.Roles.OWNER]),
        )
        self.assertEqual(
            Group.Roles.MEMBER.implied_by_group_roles,
            frozenset([Group.Roles.ADMIN, Group.Roles.MEMBER]),
        )

    def test_q_admin(self) -> None:
        member = self.playground.create_user("member")
        admin = self.playground.create_user("group_admin")
        scope_admin = self.playground.create_user("scope_admin")
        self.playground.create_group_role(
            self.scenario.scope, Scope.Roles.OWNER, users=[scope_admin]
        )
        workspace_admin = self.playground.create_user("workspace_admin")
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.OWNER,
            users=[workspace_admin],
        )

        group = self.playground.create_group("group")
        with context.disable_permission_checks():
            group.add_user(member)
            group.add_user(admin, Group.Roles.ADMIN)

        for pc, expected in (
            (PermissionContext.for_ui_anonymous(), False),
            (PermissionContext.for_ui_user(member), False),
            (PermissionContext.for_ui_user(admin), True),
            (PermissionContext.for_ui_user(workspace_admin), False),
            (PermissionContext.for_ui_user(scope_admin), True),
        ):
            with self.subTest(pc=str(pc)):
                expected_qs = [group] if expected else []
                self.assertQuerySetEqual(
                    Group.objects.filter(pk=group.pk)
                    .filter(Group.Roles.ADMIN.q(pc))
                    .distinct(),
                    expected_qs,
                )

        with context.disable_permission_checks():
            group.workspace = self.scenario.workspace
            group.save()

        with self.subTest(pc=str(pc), workspace=True):
            self.assertQuerySetEqual(
                Group.objects.filter(pk=group.pk)
                .filter(
                    Group.Roles.ADMIN.q(
                        PermissionContext.for_ui_user(workspace_admin)
                    )
                )
                .distinct(),
                [group],
            )

    def test_q_member(self) -> None:
        member = self.playground.create_user("member")
        admin = self.playground.create_user("group_admin")
        scope_admin = self.playground.create_user("scope_admin")
        self.playground.create_group_role(
            self.scenario.scope, Scope.Roles.OWNER, users=[scope_admin]
        )
        workspace_admin = self.playground.create_user("workspace_admin")
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.OWNER,
            users=[workspace_admin],
        )

        group = self.playground.create_group("group")
        with context.disable_permission_checks():
            group.add_user(member)
            group.add_user(admin, Group.Roles.ADMIN)

        for pc, expected in (
            (PermissionContext.for_ui_anonymous(), False),
            (PermissionContext.for_ui_user(member), True),
            (PermissionContext.for_ui_user(admin), True),
            (PermissionContext.for_ui_user(workspace_admin), False),
            (PermissionContext.for_ui_user(scope_admin), True),
        ):
            with self.subTest(pc=str(pc)):
                expected_qs = [group] if expected else []
                self.assertQuerySetEqual(
                    Group.objects.filter(pk=group.pk)
                    .filter(Group.Roles.MEMBER.q(pc))
                    .distinct(),
                    expected_qs,
                )

        with context.disable_permission_checks():
            group.workspace = self.scenario.workspace
            group.save()

        with self.subTest(pc=str(pc), workspace=True):
            self.assertQuerySetEqual(
                Group.objects.filter(pk=group.pk)
                .filter(
                    Group.Roles.MEMBER.q(
                        PermissionContext.for_ui_user(workspace_admin)
                    )
                )
                .distinct(),
                [group],
            )


class GroupManagerTests(TestCase):
    """Test for GroupManager class."""

    scenario = scenarios.DefaultContext()
    scope1: ClassVar[Scope]
    scope2: ClassVar[Scope]
    group1: ClassVar[Group]
    group2: ClassVar[Group]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.scope1 = cls.playground.get_or_create_scope("scope1")
        cls.group1 = cls.playground.create_group("group", scope=cls.scope1)
        cls.scope2 = cls.playground.get_or_create_scope("scope2")
        cls.group2 = cls.playground.create_group("group", scope=cls.scope2)

    def test_get_by_name(self) -> None:
        wsgroup = self.playground.create_group(
            "group", workspace=self.scenario.workspace
        )
        for arg, result in (
            ("scope1/group", self.group1),
            ("scope2/group", self.group2),
            ("debusine/System/group", wsgroup),
        ):
            with self.subTest(arg=arg):
                self.assertEqual(Group.objects.get_by_name(arg), result)

    def test_get_by_name_fail_lookup(self) -> None:
        for arg, exc_class in (
            ("scope1/fail", Group.DoesNotExist),
            ("scope1/fail/group", Group.DoesNotExist),
            ("scope2:group", ValueError),
        ):
            with self.subTest(arg=arg), self.assertRaises(exc_class):
                Group.objects.get_by_name(arg)

    def test_create_ephemeral(self) -> None:
        """Test create_ephemeral."""
        self.scenario.set_current()
        user = self.playground.get_default_user()
        group = Group.objects.create_ephemeral(
            self.scenario.workspace, name="test", owner=user
        )
        self.assertEqual(group.scope, self.scenario.scope)
        self.assertEqual(group.workspace, self.scenario.workspace)
        self.assertEqual(group.name, "test")
        self.assertQuerySetEqual(group.users.all(), [user])

        membership = GroupMembership.objects.get(group=group, user=user)
        self.assertEqual(membership.role, Group.Roles.ADMIN)

        audit_logs = list(GroupAuditLog.objects.order_by("created_at"))
        self.assertEqual(len(audit_logs), 2)
        # Group creation audit log
        audit_log = audit_logs[0]
        self.assertEqual(audit_log.group, group)
        self.assertEqual(audit_log.actor, self.scenario.user)
        changes_add = auth.GroupAuditLogCreated.create(group)
        changes_add.actor = self.scenario.user.username
        self.assertEqual(audit_log.changes, changes_add.model_dump())
        # Admin added audit log
        audit_log = audit_logs[1]
        self.assertEqual(audit_log.group, group)
        self.assertEqual(audit_log.actor, self.scenario.user)
        changes_memb = auth.GroupAuditLogMemberAdded(
            user=self.scenario.user.username,
            role=Group.Roles.ADMIN,
            actor=self.scenario.user.username,
        )
        self.assertEqual(audit_log.changes, changes_memb.model_dump())

    def test_unused_ephemeral_unused(self) -> None:
        """Test unused_ephemeral."""
        self.assertQuerySetEqual(Group.objects.unused_ephemeral(), [])

        with context.disable_permission_checks():
            self.group1.ephemeral = True
            self.group1.save()
        self.assertQuerySetEqual(
            Group.objects.unused_ephemeral(), [self.group1]
        )

    def test_unused_ephemeral_used_by_scope(self) -> None:
        """Test unused_ephemeral for a group used by a Scope."""
        with context.disable_permission_checks():
            self.group1.ephemeral = True
            self.group1.save()
        self.assertQuerySetEqual(
            Group.objects.unused_ephemeral(), [self.group1]
        )

        self.group1.assign_role(self.scope1, Scope.Roles.OWNER)
        self.assertQuerySetEqual(Group.objects.unused_ephemeral(), [])

    def test_unused_ephemeral_used_by_workspace(self) -> None:
        """Test unused_ephemeral for a group used by a Workspace."""
        with context.disable_permission_checks():
            self.group1.ephemeral = True
            self.group1.save()
        self.assertQuerySetEqual(
            Group.objects.unused_ephemeral(), [self.group1]
        )

        workspace = self.playground.create_workspace(scope=self.scope1)
        self.group1.assign_role(workspace, workspace.Roles.OWNER)
        self.assertQuerySetEqual(Group.objects.unused_ephemeral(), [])

    def test_unused_ephemeral_used_by_asset(self) -> None:
        """Test unused_ephemeral for a group used by an Asset."""
        with context.disable_permission_checks():
            self.group1.ephemeral = True
            self.group1.save()
        self.assertQuerySetEqual(
            Group.objects.unused_ephemeral(), [self.group1]
        )

        workspace = self.playground.create_workspace(scope=self.scope1)
        asset = self.playground.create_signing_key_asset(workspace=workspace)
        self.group1.assign_role(asset, asset.Roles.OWNER)
        self.assertQuerySetEqual(Group.objects.unused_ephemeral(), [])

    def test_unused_ephemeral_used_by_asset_usage(self) -> None:
        """Test unused_ephemeral for a group used by an AssetUsage."""
        with context.disable_permission_checks():
            self.group1.ephemeral = True
            self.group1.save()
        self.assertQuerySetEqual(
            Group.objects.unused_ephemeral(), [self.group1]
        )

        workspace = self.playground.create_workspace(scope=self.scope1)
        asset = self.playground.create_signing_key_asset(workspace=workspace)
        asset_usage = self.playground.create_asset_usage(
            asset, workspace=workspace
        )
        self.group1.assign_role(asset_usage, asset_usage.Roles.SIGNER)
        self.assertQuerySetEqual(Group.objects.unused_ephemeral(), [])


class GroupTests(TestCase):
    """Test for Group class."""

    scenario = scenarios.DefaultContext()

    group: ClassVar[Group]
    other_scope: ClassVar[Scope]
    other_group: ClassVar[Group]
    workspace_group: ClassVar[Group]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.group = cls.playground.create_group("group")
        cls.other_scope = cls.playground.get_or_create_scope(name="other")
        cls.other_group = cls.playground.create_group(
            "other", scope=cls.other_scope
        )
        cls.workspace_group = cls.playground.create_group(
            "group", workspace=cls.scenario.workspace
        )

    def assertRole(
        self, group: Group, user: User, role: Group.Roles | None
    ) -> None:
        """
        Ensure the user has the given role in the group.

        Set role to None to ensure the user is not a member of the group.
        """
        membership: GroupMembership | None
        try:
            membership = GroupMembership.objects.get(user=user, group=group)
        except GroupMembership.DoesNotExist:
            membership = None
        if role is None:
            if membership is None:
                pass
            else:
                self.fail(
                    f"{user} unexpectedly has role {membership.role} in {group}"
                )
        else:
            if membership is None:
                self.fail(f"{user} is unexpectedly not a member of {group}")
            else:
                self.assertEqual(membership.role, role)

    @contextlib.contextmanager
    def assertGroupAuditLog(
        self,
    ) -> Generator[list[auth.GroupAuditLogEntry]]:
        """
        Ensure add_audit_log is called.

        Generates a list populated with the messages logged.
        """
        changes: list[auth.GroupAuditLogEntry] = []
        with mock.patch(
            "debusine.db.models.auth.Group.add_audit_log"
        ) as add_audit_log:
            yield changes
        add_audit_log.assert_called()
        changes += [call.args[0] for call in add_audit_log.call_args_list]

    def test_assertRole(self) -> None:
        """Test assertRole."""
        self.scenario.set_current()
        self.assertRole(self.group, self.scenario.user, None)
        with self.assertRaisesRegex(
            AssertionError, "is unexpectedly not a member of debusine/group"
        ):
            self.assertRole(self.group, self.scenario.user, Group.Roles.MEMBER)
        self.group.add_user(self.scenario.user, Group.Roles.ADMIN)
        self.assertRole(self.group, self.scenario.user, Group.Roles.ADMIN)
        with self.assertRaisesRegex(
            AssertionError, re.escape(f"'admin' != {Group.Roles.MEMBER!r}")
        ):
            self.assertRole(self.group, self.scenario.user, Group.Roles.MEMBER)
        with self.assertRaisesRegex(
            AssertionError,
            "playground unexpectedly has role admin in debusine/group",
        ):
            self.assertRole(self.group, self.scenario.user, None)

    def test_str(self) -> None:
        self.assertEqual(str(self.group), f"{self.scenario.scope.name}/group")

    def test_str_workspaced(self) -> None:
        self.group.workspace = self.playground.create_workspace(name="other")
        self.assertEqual(
            str(self.group), f"{self.scenario.scope.name}/other/group"
        )

    def test_name_without_scope(self) -> None:
        self.assertEqual(self.group.name_without_scope, "group")

    def test_name_without_scope_workspaced(self) -> None:
        self.group.workspace = self.playground.create_workspace(name="other")
        self.assertEqual(self.group.name_without_scope, "other/group")

    def test_get_absolute_url(self) -> None:
        """Test the get_absolute_url method."""
        for group, expected in (
            (self.group, "/debusine/-/groups/group/"),
            (self.other_group, "/other/-/groups/other/"),
            (self.workspace_group, "/debusine/System/groups/group/"),
        ):
            with self.subTest(group=repr(group)):
                self.assertEqual(group.get_absolute_url(), expected)

    def test_get_absolute_url_audit_log(self) -> None:
        """Test the get_absolute_url_audit_log method."""
        for group, expected in (
            (self.group, "/debusine/-/groups/group/audit-log/"),
            (self.other_group, "/other/-/groups/other/audit-log/"),
            (self.workspace_group, "/debusine/System/groups/group/audit-log/"),
        ):
            with self.subTest(group=repr(group)):
                self.assertEqual(group.get_absolute_url_audit_log(), expected)

    def test_ui(self) -> None:
        from debusine.web.views.ui.auth import GroupUI

        request = RequestFactory().get("/")
        ui = self.group.ui(request)
        self.assertIsInstance(ui, GroupUI)

    def test_defaults(self) -> None:
        """Test default values."""
        self.assertFalse(self.group.ephemeral)

    @mock.patch("debusine.db.models.Group.add_audit_log")
    def test_unique(self, _: mock.MagicMock) -> None:
        scope2 = self.playground.get_or_create_scope("scope2")
        group2 = self.playground.create_group(self.group.name, scope=scope2)
        self.assertNotEqual(self.group.pk, group2.pk)

        with self.assertRaisesRegex(
            IntegrityError,
            # Only match constraint name to support non-english locales
            r"db_group_unique_name_scope_workspace",
        ):
            Group.objects.create(name=self.group.name, scope=self.group.scope)

    @mock.patch("debusine.db.models.Group.add_audit_log")
    def test_unique_with_workspace(self, _: mock.MagicMock) -> None:
        other_scope = self.playground.get_or_create_scope("scope2")
        workspace = self.playground.create_workspace(name="other")
        group2 = self.playground.create_group(
            self.group.name, workspace=workspace
        )
        self.assertNotEqual(self.group.pk, group2.pk)
        group3 = self.playground.create_group(
            self.group.name, scope=other_scope, workspace=workspace
        )
        self.assertNotEqual(group3.pk, group2.pk)

        with self.assertRaisesRegex(
            IntegrityError,
            # Only match constraint name to support non-english locales
            r"db_group_unique_name_scope_workspace",
        ):
            Group.objects.create(
                name=self.group.name,
                scope=self.group.scope,
                workspace=self.group.workspace,
            )

    def test_assign_user(self) -> None:
        """Test assigning users to groups."""
        user1 = User.objects.create_user(
            username="user1",
            email="user1@debian.org",
        )
        user2 = User.objects.create_user(
            username="user2",
            email="user2@debian.org",
        )

        self.playground.add_user(self.group, user1)

        self.assertQuerySetEqual(self.group.users.all(), [user1])
        self.assertQuerySetEqual(user1.groups.all(), [])
        self.assertQuerySetEqual(user1.debusine_groups.all(), [self.group])
        self.assertEqual(
            user1.group_memberships.get(group=self.group).role,
            Group.Roles.MEMBER,
        )

        self.playground.add_user(self.group, user2)

        self.assertQuerySetEqual(
            self.group.users.all(), [user1, user2], ordered=False
        )
        self.assertQuerySetEqual(user2.groups.all(), [])
        self.assertQuerySetEqual(user2.debusine_groups.all(), [self.group])

    def test_assign_role(self) -> None:
        """Test assign_role."""
        for resource, role in (
            (self.scenario.scope, Scope.Roles.OWNER),
            (self.scenario.scope, "owner"),
            (self.scenario.workspace, Workspace.Roles.OWNER),
            (self.scenario.workspace, "owner"),
        ):
            with self.subTest(resource=repr(resource), role=repr(role)):
                assigned = self.group.assign_role(resource, role)
                # TODO: we can remove the type ignores once we find a way to
                # generically type the role assignment models
                self.assertEqual(
                    assigned.resource,  # type: ignore[attr-defined]
                    resource,
                )
                self.assertEqual(
                    assigned.group,  # type: ignore[attr-defined]
                    self.group,
                )
                self.assertEqual(
                    assigned.role,  # type: ignore[attr-defined]
                    role,
                )

    def test_assign_role_not_a_resource(self) -> None:
        """Try assigning role to something which is not a resource."""
        token = self.playground.create_user_token()

        with self.assertRaisesRegex(
            NotImplementedError, r"Cannot get scope for Token object"
        ):
            self.group.assign_role(token, "owner")

    def test_assign_role_wrong_scope(self) -> None:
        """Try assigning role to a resource in a different scope."""
        scope2 = self.playground.get_or_create_scope("scope2")
        with self.assertRaisesRegex(
            ValueError, r"Scope 'scope2' is not in scope debusine"
        ):
            self.group.assign_role(scope2, "owner")

    def test_assign_role_wrong_role_enum(self) -> None:
        """Try assigning a role using the wrong enum."""
        with self.assertRaisesRegex(
            TypeError,
            re.escape(
                f"{Scope.Roles.OWNER!r} cannot be converted to"
                " <enum 'WorkspaceRoles'>"
            ),
        ):
            self.group.assign_role(self.scenario.workspace, Scope.Roles.OWNER)

    def test_assign_role_wrong_role_name(self) -> None:
        """Try assigning a role using the wrong enum."""
        with self.assertRaisesRegex(
            ValueError, r"'does-not-exist' is not a valid WorkspaceRoles"
        ):
            self.group.assign_role(self.scenario.workspace, "does-not-exist")

    def test_get_user_role(self) -> None:
        """Ensure get_user_role returns the user's role."""
        self.playground.add_user(self.group, self.scenario.user)
        self.assertEqual(
            self.group.get_user_role(self.scenario.user), Group.Roles.MEMBER
        )

        GroupMembership.objects.create(
            user=self.scenario.user,
            group=self.other_group,
            role=Group.Roles.ADMIN,
        )
        self.assertEqual(
            self.other_group.get_user_role(self.scenario.user),
            Group.Roles.ADMIN,
        )

    def test_get_user_role_not_member(self) -> None:
        """get_user_role raises ValueError if the user is not a member."""
        with self.assertRaisesRegex(
            ValueError,
            r"User playground is not a member of group debusine/group",
        ):
            self.group.get_user_role(self.scenario.user)

    def test_set_user_role(self) -> None:
        """Set the role of a user on a group."""
        self.scenario.set_current()
        self.group.add_user(self.scenario.user)
        self.assertEqual(
            self.group.get_user_role(self.scenario.user), Group.Roles.MEMBER
        )
        with self.assertGroupAuditLog() as changes:
            self.group.set_user_role(self.scenario.user, Group.Roles.ADMIN)
        self.assertEqual(
            self.group.get_user_role(self.scenario.user), Group.Roles.ADMIN
        )
        self.assertEqual(
            changes,
            [
                auth.GroupAuditLogMemberRoleChanged(
                    user=self.scenario.user.username, role=Group.Roles.ADMIN
                )
            ],
        )

        with self.assertGroupAuditLog() as changes:
            self.group.set_user_role(self.scenario.user, Group.Roles.MEMBER)
        self.assertEqual(
            self.group.get_user_role(self.scenario.user), Group.Roles.MEMBER
        )
        self.assertEqual(
            changes,
            [
                auth.GroupAuditLogMemberRoleChanged(
                    user=self.scenario.user.username, role=Group.Roles.MEMBER
                )
            ],
        )

    def test_set_user_role_not_a_member(self) -> None:
        """Set the role of a user not a member of the group."""
        with self.assertRaisesRegex(
            ValueError,
            r"User playground is not a member of group debusine/group",
        ):
            self.group.set_user_role(self.scenario.user, Group.Roles.ADMIN)

    def test_add_user(self) -> None:
        """Add a nonexisting user to a group."""
        self.assertRole(self.group, self.scenario.user, None)
        with self.assertGroupAuditLog() as changes:
            self.assertIsInstance(
                self.group.add_user(self.scenario.user), GroupMembership
            )
        self.assertRole(self.group, self.scenario.user, Group.Roles.MEMBER)
        self.assertEqual(
            changes,
            [
                auth.GroupAuditLogMemberAdded(
                    user=self.scenario.user.username, role=Group.Roles.MEMBER
                )
            ],
        )

    def test_add_user_role(self) -> None:
        """Add a user to a group with a custom role."""
        self.assertRole(self.group, self.scenario.user, None)
        with self.assertGroupAuditLog() as changes:
            self.group.add_user(self.scenario.user, Group.Roles.ADMIN)
        self.assertRole(self.group, self.scenario.user, Group.Roles.ADMIN)

        self.assertEqual(
            changes,
            [
                auth.GroupAuditLogMemberAdded(
                    user=self.scenario.user.username, role=Group.Roles.ADMIN
                )
            ],
        )

    def test_add_user_already_member(self) -> None:
        """Cannot add a user that is already a member."""
        self.scenario.set_current()
        self.group.add_user(self.scenario.user)
        self.assertRole(self.group, self.scenario.user, Group.Roles.MEMBER)
        with self.assertRaisesRegex(
            ValueError,
            r"playground is already a member of group debusine/group",
        ):
            self.group.add_user(self.scenario.user, Group.Roles.ADMIN)
        self.assertRole(self.group, self.scenario.user, Group.Roles.MEMBER)

    def test_remove_user(self) -> None:
        """Remove a user from a group."""
        self.playground.add_user(self.group, self.scenario.user)
        self.assertRole(self.group, self.scenario.user, Group.Roles.MEMBER)
        with self.assertGroupAuditLog() as changes:
            self.group.remove_user(self.scenario.user)
        self.assertRole(self.group, self.scenario.user, None)
        self.assertEqual(
            changes,
            [auth.GroupAuditLogMemberRemoved(user=self.scenario.user.username)],
        )

    def test_remove_user_not_member(self) -> None:
        """Removing a user not in the group fails."""
        self.assertRole(self.group, self.scenario.user, None)
        with self.assertRaisesRegex(
            ValueError,
            r"User playground is not a member of group debusine/group",
        ):
            self.group.remove_user(self.scenario.user)
        self.assertRole(self.group, self.scenario.user, None)

    def test_add_audit_log_no_user(self) -> None:
        """Ensure add_audit_log fails if there is no user in context."""
        context.set_permission_context_scoped(
            self.scenario.scope, PermissionContext.for_ui_anonymous()
        )

        with self.assertRaisesRegex(
            ContextConsistencyError, r"add_audit_log requires a valid user"
        ):
            self.group.add_audit_log(
                auth.GroupAuditLogMemberAdded(
                    user="test", role=Group.Roles.MEMBER
                )
            )

    def test_add_audit_log(self) -> None:
        """Ensure add_audit_log fails if there is no user in context."""
        self.scenario.set_current()
        changes = auth.GroupAuditLogMemberAdded(
            user=self.scenario.user.username, role=Group.Roles.ADMIN
        )
        entry = self.group.add_audit_log(changes.model_copy())
        assert entry is not None
        changes.actor = self.scenario.user.username
        self.assertEqual(entry.group, self.group)
        self.assertEqual(entry.actor, self.scenario.user)
        self.assertEqual(entry.changes, changes.model_dump())
        self.assertIsInstance(
            entry.changes_model, auth.GroupAuditLogMemberAdded
        )
        self.assertEqual(entry.changes_model, changes)

    def test_can_display_anonymous(self) -> None:
        """Deny can_display for anonymous users."""
        self.assertPermission(
            "can_display",
            users=AnonymousUser(),
            denied=[self.group, self.other_group, self.workspace_group],
        )

    def test_can_display_workers(self) -> None:
        """Deny can_display for workers."""
        GroupMembership.objects.create(
            group=self.group, user=self.scenario.user, role=Group.Roles.ADMIN
        )

        self.assertPermission(
            "can_display",
            users=[None],
            denied=[self.group, self.other_group, self.workspace_group],
            token=self.playground.create_worker_token(),
        )

    def test_can_display(self) -> None:
        """can_display is True for any user."""
        owner_group = self.playground.create_group_role(
            self.scenario.scope, "owner", [self.scenario.user]
        )
        admin = self.playground.create_user("admin")
        self.playground.add_user(self.group, admin, Group.Roles.ADMIN)

        member = self.playground.create_user("member")
        self.playground.add_user(self.group, member)

        other = self.playground.create_user("other")

        self.assertPermission(
            "can_display",
            users=[self.scenario.user, admin, member, other],
            allowed=[
                self.group,
                owner_group,
                self.other_group,
                self.workspace_group,
            ],
        )

    def test_can_display_audit_log_anonymous(self) -> None:
        """Deny can_display_audit_log for anonymous users."""
        self.assertPermission(
            "can_display_audit_log",
            users=AnonymousUser(),
            denied=[self.group, self.other_group, self.workspace_group],
        )

    def test_can_display_audit_log_workers(self) -> None:
        """Deny can_display_audit_log for workers."""
        GroupMembership.objects.create(
            group=self.group, user=self.scenario.user, role=Group.Roles.ADMIN
        )

        self.assertPermission(
            "can_display_audit_log",
            users=[None],
            denied=[self.group, self.other_group, self.workspace_group],
            token=self.playground.create_worker_token(),
        )

    def test_can_display_audit_log(self) -> None:
        """can_display_audit_log is True for any user."""
        owner_group = self.playground.create_group_role(
            self.scenario.scope, "owner", [self.scenario.user]
        )
        admin = self.playground.create_user("admin")
        GroupMembership.objects.create(
            group=self.group, user=admin, role=Group.Roles.ADMIN
        )

        member = self.playground.create_user("member")
        self.group.users.add(member)

        other = self.playground.create_user("other")

        self.assertPermission(
            "can_display_audit_log",
            users=[self.scenario.user, admin, member, other],
            allowed=[
                self.group,
                owner_group,
                self.other_group,
                self.workspace_group,
            ],
        )

    def test_can_manage_anonymous(self) -> None:
        self.assertPermission(
            "can_manage",
            users=AnonymousUser(),
            denied=[self.group, self.other_group, self.workspace_group],
        )

    def test_can_manage_workers(self) -> None:
        GroupMembership.objects.create(
            group=self.group, user=self.scenario.user, role=Group.Roles.ADMIN
        )

        self.assertPermission(
            "can_manage",
            users=[None],
            denied=[self.group, self.other_group, self.workspace_group],
            token=self.playground.create_worker_token(),
        )

    def test_can_manage_scope_owner(self) -> None:
        owner_group = self.playground.create_group_role(
            self.scenario.scope, "owner", [self.scenario.user]
        )
        self.assertPermission(
            "can_manage",
            users=self.scenario.user,
            allowed=[self.group, owner_group, self.workspace_group],
            denied=[self.other_group],
        )

    def test_can_manage_scope_owner_context_shortcut(self) -> None:
        self.playground.create_group_role(
            self.scenario.scope, "owner", [self.scenario.user]
        )
        self.scenario.set_current()
        with self.assertNumQueries(0):
            self.assertTrue(self.group.can_manage(context.pc))

    def test_can_manage_workspace_owner(self) -> None:
        owner_group = self.playground.create_group_role(
            self.scenario.workspace, "owner", [self.scenario.user]
        )
        self.assertPermission(
            "can_manage",
            users=self.scenario.user,
            allowed=[owner_group, self.workspace_group],
            denied=[self.group, self.other_group],
        )

    def test_can_manage_workspace_owner_context_shortcut(self) -> None:
        self.playground.create_group_role(
            self.scenario.workspace, "owner", [self.scenario.user]
        )
        self.scenario.set_current()
        with self.assertNumQueries(0):
            self.assertTrue(self.workspace_group.can_manage(context.pc))

    def test_can_manage_group_member(self) -> None:
        self.playground.add_user(self.group, self.scenario.user)
        self.assertPermission(
            "can_manage",
            users=(None, AnonymousUser(), self.scenario.user),
            denied=[self.group, self.other_group, self.workspace_group],
            token=self.playground.create_worker_token(),
        )

    def test_can_manage_group_admin(self) -> None:
        GroupMembership.objects.create(
            group=self.group, user=self.scenario.user, role=Group.Roles.ADMIN
        )
        self.assertPermission(
            "can_manage",
            users=self.scenario.user,
            allowed=self.group,
            denied=[self.other_group, self.workspace_group],
        )

    def test_can_grant_anonymous(self) -> None:
        self.assertPermission(
            "can_grant",
            users=AnonymousUser(),
            denied=[self.group, self.other_group, self.workspace_group],
        )

    def test_can_grant_workers(self) -> None:
        GroupMembership.objects.create(
            group=self.group, user=self.scenario.user, role=Group.Roles.ADMIN
        )

        self.assertPermission(
            "can_grant",
            users=[None],
            denied=[self.group, self.other_group, self.workspace_group],
            token=self.playground.create_worker_token(),
        )

    def test_can_grant_scope_owner(self) -> None:
        owner_group = self.playground.create_group_role(
            self.scenario.scope, "owner", [self.scenario.user]
        )
        self.assertPermission(
            "can_grant",
            users=self.scenario.user,
            allowed=[self.group, owner_group, self.workspace_group],
            denied=[self.other_group],
        )

    def test_can_grant_scope_owner_context_shortcut(self) -> None:
        self.playground.create_group_role(
            self.scenario.scope, "owner", [self.scenario.user]
        )
        self.scenario.set_current()
        with self.assertNumQueries(0):
            self.assertTrue(self.group.can_grant(context.pc))

    def test_can_grant_workspace_owner(self) -> None:
        owner_group = self.playground.create_group_role(
            self.scenario.workspace, "owner", [self.scenario.user]
        )
        self.assertPermission(
            "can_grant",
            users=self.scenario.user,
            allowed=[owner_group, self.workspace_group],
            denied=[self.group, self.other_group],
        )

    def test_can_grant_workspace_owner_context_shortcut(self) -> None:
        self.playground.create_group_role(
            self.scenario.workspace, "owner", [self.scenario.user]
        )
        self.scenario.set_current()
        with self.assertNumQueries(0):
            self.assertTrue(self.workspace_group.can_grant(context.pc))

    def test_can_grant_group_member(self) -> None:
        self.playground.add_user(self.group, self.scenario.user)
        self.assertPermission(
            "can_grant",
            users=self.scenario.user,
            allowed=[self.group],
            denied=[self.other_group, self.workspace_group],
        )

    def test_can_grant_group_admin(self) -> None:
        GroupMembership.objects.create(
            group=self.group, user=self.scenario.user, role=Group.Roles.ADMIN
        )
        self.assertPermission(
            "can_grant",
            users=self.scenario.user,
            allowed=self.group,
            denied=[self.other_group, self.workspace_group],
        )


class GroupMembershipTests(TestCase):
    """Test for py:class:`GroupMembership`."""

    scenario = scenarios.DefaultContext()

    other_scope: ClassVar[Scope]

    group: ClassVar[Group]
    other_group: ClassVar[Group]
    workspace_group: ClassVar[Group]

    membership: ClassVar[GroupMembership]
    other_membership: ClassVar[GroupMembership]
    workspace_membership: ClassVar[GroupMembership]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.other_scope = cls.playground.get_or_create_scope(name="other")

        cls.group = cls.playground.create_group(
            "group", users=[cls.scenario.user]
        )
        cls.other_group = cls.playground.create_group(
            "group", scope=cls.other_scope, users=[cls.scenario.user]
        )
        cls.workspace_group = cls.playground.create_group(
            "group", workspace=cls.scenario.workspace, users=[cls.scenario.user]
        )

        cls.membership = GroupMembership.objects.get(
            user=cls.scenario.user, group=cls.group
        )
        cls.other_membership = GroupMembership.objects.get(
            user=cls.scenario.user, group=cls.other_group
        )
        cls.workspace_membership = GroupMembership.objects.get(
            user=cls.scenario.user, group=cls.workspace_group
        )

    def test_str(self) -> None:
        for membership, expected in (
            (self.membership, "debusine/group:playground"),
            (self.other_membership, "other/group:playground"),
            (
                self.workspace_membership,
                "debusine/System/group:playground",
            ),
        ):
            with self.subTest(group=repr(membership.group)):
                self.assertEqual(str(membership), expected)

    def test_get_absolute_url_update(self) -> None:
        for membership, expected in (
            (self.membership, "/debusine/-/groups/group/edit/playground/"),
            (self.other_membership, "/other/-/groups/group/edit/playground/"),
            (
                self.workspace_membership,
                "/debusine/System/groups/group/edit/playground/",
            ),
        ):
            with self.subTest(group=repr(membership.group)):
                self.assertEqual(membership.get_absolute_url_update(), expected)

    def test_get_absolute_url_remove(self) -> None:
        for membership, expected in (
            (self.membership, "/debusine/-/groups/group/remove/playground/"),
            (self.other_membership, "/other/-/groups/group/remove/playground/"),
            (
                self.workspace_membership,
                "/debusine/System/groups/group/remove/playground/",
            ),
        ):
            with self.subTest(group=repr(membership.group)):
                self.assertEqual(membership.get_absolute_url_remove(), expected)


class GroupAuditLogEntryTests(TestCase):
    """Test GroupAuditLogEntry models."""

    scenario = scenarios.DefaultContext()

    def test_deserialize_created_pre_workspace(self) -> None:
        obj = group_audit_log_entry_adapter.validate_python(
            {
                "type": GroupAuditLogEntryTypes.CREATED,
                "name": "test",
                "scope": "scope",
                "ephemeral": False,
            },
        )
        assert isinstance(obj, auth.GroupAuditLogCreated)
        self.assertEqual(obj.type, GroupAuditLogEntryTypes.CREATED)
        self.assertEqual(obj.name, "test")
        self.assertEqual(obj.scope, "scope")
        self.assertIsNone(obj.workspace)
        self.assertFalse(obj.ephemeral, False)

    def test_deserialize_updated_pre_workspace(self) -> None:
        obj = group_audit_log_entry_adapter.validate_python(
            {
                "type": GroupAuditLogEntryTypes.UPDATED,
                "name": ("test", "test1"),
                "scope": None,
                "ephemeral": (False, True),
            },
        )
        assert isinstance(obj, auth.GroupAuditLogUpdated)
        self.assertEqual(obj.type, GroupAuditLogEntryTypes.UPDATED)
        self.assertEqual(obj.name, ("test", "test1"))
        self.assertIsNone(obj.scope)
        self.assertIsNone(obj.workspace)
        self.assertEqual(obj.ephemeral, (False, True))

    def test_created_create(self) -> None:
        group = Group(name="test", scope=self.scenario.scope, ephemeral=True)
        entry = auth.GroupAuditLogCreated.create(group)
        self.assertEqual(
            entry,
            auth.GroupAuditLogCreated(
                name="test",
                scope=self.scenario.scope.name,
                workspace=None,
                ephemeral=True,
            ),
        )

    def test_created_create_workspace_scoped(self) -> None:
        group = Group(
            name="test",
            scope=self.scenario.scope,
            workspace=self.scenario.workspace,
            ephemeral=True,
        )
        entry = auth.GroupAuditLogCreated.create(group)
        self.assertEqual(
            entry,
            auth.GroupAuditLogCreated(
                name="test",
                scope=self.scenario.scope.name,
                workspace=self.scenario.workspace.name,
                ephemeral=True,
            ),
        )

    def test_created_description(self) -> None:
        """Test GroupAuditLogCreated.description."""
        for group, description in (
            (
                Group(name="test", scope=self.scenario.scope, ephemeral=True),
                "actor created ephemeral group test in scope debusine",
            ),
            (
                Group(name="test", scope=self.scenario.scope, ephemeral=False),
                "actor created group test in scope debusine",
            ),
            (
                Group(
                    name="test",
                    scope=self.scenario.scope,
                    workspace=self.scenario.workspace,
                    ephemeral=False,
                ),
                f"actor created group {self.scenario.workspace.name}/test"
                " in scope debusine",
            ),
        ):
            with self.subTest(group=repr(group)):
                entry = auth.GroupAuditLogCreated.create(group)
                entry.actor = "actor"
                self.assertEqual(
                    entry.description(),
                    description,
                )

    def test_updated_create_diff(self) -> None:
        scope1 = self.playground.get_or_create_scope("scope1")
        group1 = Group(name="test", scope=self.scenario.scope, ephemeral=False)
        group2 = Group(name="test1", scope=scope1, ephemeral=True)
        entry = auth.GroupAuditLogUpdated.create_diff(group1, group2)
        self.assertEqual(
            entry,
            auth.GroupAuditLogUpdated(
                name=("test", "test1"),
                scope=(self.scenario.scope.name, "scope1"),
                ephemeral=(False, True),
            ),
        )

    def test_updated_create_diff_workspace(self) -> None:
        group1 = Group(name="test", scope=self.scenario.scope, ephemeral=False)
        group2 = Group(
            name="test",
            scope=self.scenario.scope,
            workspace=self.scenario.workspace,
            ephemeral=False,
        )
        entry = auth.GroupAuditLogUpdated.create_diff(group1, group2)
        self.assertEqual(
            entry,
            auth.GroupAuditLogUpdated(
                name=None,
                scope=None,
                workspace=(None, self.scenario.workspace.name),
                ephemeral=None,
            ),
        )

    def test_updated_description(self) -> None:
        """Test GroupAuditLogUpdated.description."""
        group = Group(name="test", scope=self.scenario.scope, ephemeral=False)
        scope1 = self.playground.get_or_create_scope("scope1")
        for other_group, expected in (
            (
                Group(name="test1", scope=scope1, ephemeral=True),
                "actor updated name: 'test'→'test1';"
                " scope: 'debusine'→'scope1';"
                " ephemeral: False→True",
            ),
            (
                Group(name="test1", scope=self.scenario.scope, ephemeral=True),
                "actor updated name: 'test'→'test1'; ephemeral: False→True",
            ),
            (
                Group(name="test", scope=self.scenario.scope, ephemeral=True),
                "actor updated ephemeral: False→True",
            ),
            (
                Group(
                    name="test",
                    scope=self.scenario.scope,
                    workspace=self.scenario.workspace,
                ),
                "actor updated workspace:"
                f" None→{self.scenario.workspace.name!r}",
            ),
            (
                group,
                "actor updated group, no changes detected",
            ),
        ):
            with self.subTest(other_group=repr(other_group)):
                entry = auth.GroupAuditLogUpdated.create_diff(
                    group, other_group
                )
                entry.actor = "actor"
                self.assertEqual(entry.description(), expected)

    def test_member_added_description(self) -> None:
        """Test GroupAuditLogMemberAdded.description."""
        entry = auth.GroupAuditLogMemberAdded(
            actor="actor", user="test", role=Group.Roles.ADMIN
        )
        self.assertEqual(entry.description(), "actor added test as admin")

    def test_role_changed_description(self) -> None:
        """Test GroupAuditLogMemberRoleChanged.description."""
        entry = auth.GroupAuditLogMemberRoleChanged(
            actor="actor", user="test", role=Group.Roles.ADMIN
        )
        self.assertEqual(entry.description(), "actor set test's role as admin")

    def test_removed_description(self) -> None:
        """Test GroupAuditLogMemberRemoved.description."""
        entry = auth.GroupAuditLogMemberRemoved(actor="actor", user="test")
        self.assertEqual(entry.description(), "actor removed test")


class ClientEnrollTests(TestCase):
    """Tests for :py:class:`ClientEnroll`."""

    def test_fields(self) -> None:
        """Test basic model fields."""
        ce = ClientEnroll.objects.create(
            nonce="12345678", payload={"foo": "bar"}
        )
        self.assertEqual(ce.nonce, "12345678")
        self.assertEqual(ce.payload, {"foo": "bar"})

    def test_nonce_unique(self) -> None:
        """Nonce must be unique."""
        ClientEnroll.objects.create(nonce="12345678", payload={"foo": "bar"})
        with self.assertRaisesRegex(
            IntegrityError, "db_clientenroll_nonce_key"
        ):
            ClientEnroll.objects.create(
                nonce="12345678", payload={"foo": "baz"}
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the worker pool models."""

import datetime as dt
import logging
from unittest.mock import PropertyMock, patch

from django.core.exceptions import ValidationError
from django.db import IntegrityError
from django.test import RequestFactory
from django.utils import timezone

from debusine.assets import (
    HetznerProviderAccountConfiguration,
    HetznerProviderAccountCredentials,
    HetznerProviderAccountData,
)
from debusine.db.models import (
    ScopeWorkerPool,
    Token,
    WorkRequest,
    Worker,
    WorkerPool,
    WorkerPoolStatistics,
)
from debusine.server.worker_pools import (
    AWSEC2WorkerPoolSpecification,
    DummyWorkerPool,
    DummyWorkerPoolSpecification,
    ScopeWorkerPoolLimits,
    WorkerPoolLimits,
)
from debusine.test.django import TestCase


class WorkerPoolManagerTests(TestCase):
    """Tests for the WorkerPoolManager Manager."""

    def test_enabled(self) -> None:
        provider_account = self.playground.create_cloud_provider_account_asset()
        enabled = self.playground.create_worker_pool(
            name="enabled", provider_account=provider_account
        )
        self.playground.create_worker_pool(
            name="disabled", enabled=False, provider_account=provider_account
        )
        self.assertQuerySetEqual(WorkerPool.objects.enabled(), [enabled])


class WorkerPoolTests(TestCase):
    """Tests for the WorkerPool model."""

    def test_str(self) -> None:
        worker_pool = WorkerPool(id=42, name="foo")
        self.assertEqual(
            worker_pool.__str__(),
            "Id: 42 Name: foo",
        )

    def test_ui(self) -> None:
        from debusine.web.views.ui.worker_pools import WorkerPoolUI

        request = RequestFactory().get("/")
        worker_pool = WorkerPool(id=42, name="foo")
        ui = worker_pool.ui(request)
        self.assertIsInstance(ui, WorkerPoolUI)

    def test_get_absolute_url(self) -> None:
        """Test the get_absolute_url method."""
        worker_pool = self.playground.create_worker_pool()
        self.assertEqual(
            worker_pool.get_absolute_url(),
            f"/-/status/worker-pools/{worker_pool.name}/",
        )

    def test_limits_model(self) -> None:
        limits = WorkerPoolLimits(max_active_instances=10)
        worker_pool = WorkerPool(limits=limits.model_dump())
        self.assertEqual(worker_pool.limits_model, limits)

    def test_specifications_model(self) -> None:
        provider_account = self.playground.create_cloud_provider_account_asset()
        specifications = DummyWorkerPoolSpecification(features=["foo"])
        worker_pool = WorkerPool(
            specifications=specifications.model_dump(),
            provider_account=provider_account,
        )
        self.assertEqual(worker_pool.specifications_model, specifications)

    def test_specifications_model_mismatching_account(self) -> None:
        provider_account = self.playground.create_cloud_provider_account_asset(
            data=HetznerProviderAccountData(
                name="hetzner",
                configuration=HetznerProviderAccountConfiguration(
                    region_name="nbg1",
                ),
                credentials=HetznerProviderAccountCredentials(
                    api_token="secret token",
                ),
            )
        )
        specifications = AWSEC2WorkerPoolSpecification(launch_templates=[])
        worker_pool = WorkerPool(
            name="foo",
            specifications=specifications.model_dump(),
            provider_account=provider_account,
        )
        with self.assertRaisesRegex(
            ValueError,
            (
                r"specifications for worker_pool foo do not have a "
                r"provider_account with a matching provider_type\."
            ),
        ):
            worker_pool.specifications_model

    def create_pool_with_one_running_worker(
        self,
    ) -> tuple[WorkerPool, list[Worker], list[Worker]]:
        """Return a worker_pool with a mix of workers, some running."""
        worker_pool = self.playground.create_worker_pool()
        running: list[Worker] = []
        stopped: list[Worker] = []
        for id_ in range(3):
            worker = self.playground.create_worker(worker_pool=worker_pool)
            if id_ == 2:
                worker.instance_created_at = None
                worker.save()
                stopped.append(worker)
            else:
                running.append(worker)
        return worker_pool, running, stopped

    def test_workers_running(self) -> None:
        worker_pool, running, _ = self.create_pool_with_one_running_worker()
        self.assertQuerySetEqual(
            worker_pool.workers_running.order_by("id"), running
        )

    def test_workers_stopped(self) -> None:
        worker_pool, _, stopped = self.create_pool_with_one_running_worker()
        self.assertQuerySetEqual(
            worker_pool.workers_stopped.order_by("id"), stopped
        )

    def test_provider_interface(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        self.assertIsInstance(worker_pool.provider_interface, DummyWorkerPool)

    def test_launch_workers(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        # Create one pre-existing worker
        Worker.objects.create_pool_members(worker_pool=worker_pool, count=1)
        worker = worker_pool.worker_set.first()
        assert worker and worker.activation_token
        old_activation_token = worker.activation_token

        worker_pool.launch_workers(3)
        self.assertEqual(worker_pool.worker_set.count(), 3)
        self.assertEqual(
            worker_pool.workers_running.filter(
                worker_pool_data__launched=True
            ).count(),
            3,
        )

        worker.refresh_from_db()
        self.assertNotEqual(old_activation_token, worker.activation_token)
        self.assertIsNotNone(worker.activation_token.expire_at)
        self.assertTrue(worker.activation_token.enabled)

    def test_launch_workers_disabled(self) -> None:
        worker_pool = self.playground.create_worker_pool(enabled=False)
        with self.assertRaisesRegex(
            ValueError, r"Pool is disabled, refusing to launch workers."
        ):
            worker_pool.launch_workers(1)

    def test_launch_workers_integrity_error(self) -> None:
        """Exercise the branch from transaction.atomic() to exit."""
        with (
            patch.object(
                WorkerPool, "provider_interface", new_callable=PropertyMock
            ) as provider_interface,
            self.assertRaises(IntegrityError),
        ):
            provider_interface.side_effect = IntegrityError
            worker_pool = self.playground.create_worker_pool()
            worker_pool.launch_workers(1)

    def test_launch_workers_preexisting(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        # Create two pre-existing workers
        Worker.objects.create_pool_members(worker_pool=worker_pool, count=2)
        worker_pool.launch_workers(2)
        self.assertEqual(worker_pool.worker_set.count(), 2)
        self.assertEqual(worker_pool.workers_running.count(), 2)

    def test_launch_workers_disable_old_token(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        # Create one pre-existing worker
        Worker.objects.create_pool_members(worker_pool=worker_pool, count=1)
        worker = worker_pool.worker_set.first()
        assert worker and worker.activation_token
        activation_token = worker.activation_token
        self.playground.create_worker_token(worker=worker)
        assert worker.token
        worker.activation_token = None
        worker.save()
        activation_token.delete()

        worker_pool.launch_workers(3)
        self.assertEqual(worker_pool.worker_set.count(), 3)
        self.assertEqual(
            worker_pool.workers_running.filter(
                worker_pool_data__launched=True
            ).count(),
            3,
        )

        worker.refresh_from_db()
        assert worker.activation_token is not None
        self.assertIsNotNone(worker.activation_token.expire_at)
        self.assertTrue(worker.activation_token.enabled)
        self.assertFalse(worker.token.enabled)

    def test_terminate_worker_with_activation_token(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=worker_pool)
        assert worker.activation_token
        token = worker.activation_token
        self.assertTrue(token.enabled)

        worker_pool.terminate_worker(worker)

        with self.assertRaises(Token.DoesNotExist):
            token.refresh_from_db()

        worker.refresh_from_db()
        # The provider should have done this on termination:
        self.assertIsNone(worker.instance_created_at)
        self.assertIsNone(worker.activation_token)

    def test_terminate_worker_with_token(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=worker_pool)
        activation_token = worker.activation_token
        assert activation_token
        self.playground.create_worker_token(worker=worker)
        assert worker.token
        token = worker.token
        worker.activation_token = None
        worker.save()
        activation_token.delete()

        worker_pool.terminate_worker(worker)

        with self.assertRaises(Token.DoesNotExist):
            token.refresh_from_db()

        worker.refresh_from_db()
        # The provider should have done this on termination:
        self.assertIsNone(worker.instance_created_at)
        self.assertIsNone(worker.token)

    def test_terminate_worker_wrong_pool(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker()

        with self.assertRaisesRegex(
            ValueError,
            r"pool Id: \d+ Name: test cannot terminate worker for pool None",
        ):
            worker_pool.terminate_worker(worker)

        worker.refresh_from_db()
        # The provider should have done this on termination:
        self.assertIsNone(worker.instance_created_at)

    def test_terminate_worker_records_statistics(self) -> None:
        """Terminating a worker records running time statistics."""
        worker_pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=worker_pool)
        worker.instance_created_at = timezone.now() - dt.timedelta(minutes=1)
        worker.save()

        worker_pool.terminate_worker(worker)

        self.assertQuerySetEqual(
            WorkerPoolStatistics.objects.filter(
                worker_pool=worker_pool
            ).values_list("worker_pool", "worker", "runtime"),
            [(worker_pool.id, worker.id, 60)],
        )
        worker.refresh_from_db()
        self.assertIsNone(worker.instance_created_at)

    def test_terminate_worker_no_instance_created_at_no_statistics(
        self,
    ) -> None:
        """If instance_created_at is null, terminate_worker skips statistics."""
        worker_pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=worker_pool)
        worker.instance_created_at = None
        worker.save()

        worker_pool.terminate_worker(worker)

        self.assertFalse(
            WorkerPoolStatistics.objects.filter(
                worker_pool=worker_pool
            ).exists()
        )

    def test_terminate_worker_deassigns_running_work_request(self) -> None:
        """Terminate a worker deassigns a running work request, if possible."""
        worker_pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=worker_pool)
        work_request_completed = self.playground.create_worker_task(
            result=WorkRequest.Results.SUCCESS, worker=worker
        )
        work_request_running = self.playground.create_worker_task(
            mark_running=True, worker=worker
        )
        work_request_pending = self.playground.create_worker_task(
            mark_pending=True
        )
        work_request_pending.worker = worker
        work_request_pending.save()

        with self.assertNoLogs(
            logger="debusine.db.models.worker_pools", level=logging.DEBUG
        ):
            worker_pool.terminate_worker(worker)

        # The running work request was de-assigned.
        work_request_running.refresh_from_db()
        self.assertEqual(
            work_request_running.status, WorkRequest.Statuses.PENDING
        )
        self.assertFalse(hasattr(work_request_running, "superseded"))

        # the pending work request was de-assigned
        work_request_pending.refresh_from_db()
        self.assertEqual(
            work_request_pending.status, WorkRequest.Statuses.PENDING
        )
        self.assertIsNone(work_request_pending.worker)
        self.assertFalse(hasattr(work_request_pending, "superseded"))

        # the completed work request was left alone.
        work_request_completed.refresh_from_db()
        self.assertEqual(
            work_request_completed.status, WorkRequest.Statuses.COMPLETED
        )
        self.assertEqual(work_request_completed.worker, worker)
        self.assertFalse(hasattr(work_request_completed, "superseded"))

    def test_terminate_worker_retries_running_work_request(self) -> None:
        """Terminate a worker retries a running work request, if needed."""
        worker_pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=worker_pool)
        work_request = self.playground.create_worker_task(
            mark_running=True, worker=worker
        )
        self.playground.create_build_log_artifact(work_request=work_request)

        with self.assertNoLogs(
            logger="debusine.db.models.worker_pools", level=logging.DEBUG
        ):
            worker_pool.terminate_worker(worker)

        # The running work request was retried, though not yet assigned to a
        # worker.
        work_request.refresh_from_db()
        self.assertEqual(work_request.status, WorkRequest.Statuses.ABORTED)
        self.assertTrue(hasattr(work_request, "superseded"))
        self.assertIsNone(work_request.superseded.worker)
        self.assertEqual(work_request.superseded.workflow_data.retry_count, 1)

    def test_terminate_worker_cannot_retry_running_work_request(self) -> None:
        """If a running work request cannot be retried, it is logged."""
        worker_pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=worker_pool)
        work_request_running = self.playground.create_worker_task(worker=worker)
        work_request_running.task_data = {"invalid": True}
        work_request_running.configured_task_data = {"invalid": True}
        work_request_running.save()
        self.playground.advance_work_request(
            work_request_running, mark_running=True
        )
        self.playground.create_build_log_artifact(
            work_request=work_request_running
        )

        with self.assertLogsContains(
            "Cannot retry previously-running work request: "
            "Task dependencies cannot be satisfied",
            logger="debusine.db.models.worker_pools",
            level=logging.DEBUG,
        ):
            worker_pool.terminate_worker(worker)

        # The running work request was aborted but could not be retried.
        work_request_running.refresh_from_db()
        self.assertEqual(
            work_request_running.status, WorkRequest.Statuses.ABORTED
        )
        self.assertFalse(hasattr(work_request_running, "superseded"))

    def test_clean_valid(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        worker_pool.clean()

    def test_clean_specifications(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        worker_pool.specifications = {
            "provider_type": "dummy",
            "extra_key": "unknown",
        }
        with self.assertRaisesRegex(
            ValidationError,
            r"validation errors? for DummyWorkerPoolSpecification",
        ):
            worker_pool.clean()

    def test_clean_limits(self) -> None:
        worker_pool = self.playground.create_worker_pool()
        worker_pool.limits = {"extra_key": "unknown"}
        with self.assertRaisesRegex(
            ValidationError, r"validation errors? for WorkerPoolLimits"
        ):
            worker_pool.clean()


class ScopeWorkerPoolTests(TestCase):
    """Tests for the ScopeWorkerPool model."""

    def test_str(self) -> None:
        worker_pool = WorkerPool(name="foo")
        scope = self.playground.get_default_scope()
        scope_worker_pool = ScopeWorkerPool(
            id=42, worker_pool=worker_pool, scope=scope
        )
        self.assertEqual(
            scope_worker_pool.__str__(),
            f"Id: 42 WorkerPool: foo Scope: {scope.name}",
        )

    def test_limits_model(self) -> None:
        limits = ScopeWorkerPoolLimits(target_max_seconds_per_month=99)
        scope_worker_pool = ScopeWorkerPool(limits=limits.model_dump())
        self.assertEqual(scope_worker_pool.limits_model, limits)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test for the Config class."""

import io
from configparser import ConfigParser
from pathlib import Path
from typing import Any, override

from debusine.client.config import ConfigHandler
from debusine.test import TestCase


class ConfigHandlerTests(TestCase):
    """Tests for ConfigHandler class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.stderr = io.StringIO()

    @staticmethod
    def valid_configuration() -> ConfigParser:
        """Return a valid configuration."""
        configuration = ConfigParser()
        configuration['General'] = {'default-server': 'debian'}
        configuration['server:debian'] = {
            'api-url': 'https://debusine.debian.org/api',
            'scope': 'debian',
            'token': 'the-token-for-debian',
        }
        configuration['server:kali'] = {
            'api-url': 'https://debusine.kali.org/api',
            'scope': 'kali',
            'token': 'the-token-for-kali',
        }

        return configuration

    def write_configuration(self, configuration: ConfigParser) -> Path:
        """Return filename with the configuration."""
        temp_file = self.create_temporary_file(suffix=".ini")

        with open(temp_file, "w") as fp:
            configuration.write(fp)

        return temp_file

    def build_config_handler(
        self, config_parser: ConfigParser, **config_kwargs: Any
    ) -> ConfigHandler:
        """Return ConfigHandler for config_parser."""
        temporary_config = self.write_configuration(config_parser)

        return ConfigHandler(
            config_file_path=temporary_config,
            stderr=self.stderr,
            **config_kwargs,
        )

    def test_default_server_name(self) -> None:
        """ConfigHandler._default_server_name() returns expected name."""
        config = self.build_config_handler(self.valid_configuration())
        self.assertEqual(
            config._default_server_name(), config["General"]["default-server"]
        )

    def test_default_server_name_missing_general(self) -> None:
        """ConfigHandler._default_server_name() cannot read [General]."""
        config = self.build_config_handler(ConfigParser())

        with self.assertRaisesSystemExit(3):
            config._default_server_name()

        self.assertEqual(
            self.stderr.getvalue(),
            f'[General] section and default-server key must exist in '
            f'{config._config_file_path} to use the default server. Add '
            f'them or specify the server in the command line.\n',
        )

    def test_default_server_name_missing_key(self) -> None:
        """ConfigHandler._default_server_name() default-server key missing."""
        config_parser = ConfigParser()
        config_parser.add_section('General')

        config = self.build_config_handler(config_parser)

        with self.assertRaisesSystemExit(3):
            config._default_server_name()

        self.assertEqual(
            self.stderr.getvalue(),
            f'default-server key must exist in [General] in '
            f'{config._config_file_path} . Add it or specify the server'
            f' in the command line.\n',
        )

    def test_server(self) -> None:
        """ConfigHandler._server_configuration(server) returns the config."""
        config = self.build_config_handler(self.valid_configuration())

        server_name = "kali"

        server_config = self.valid_configuration()[f"server:{server_name}"]

        server = config._server_configuration(server_name)

        self.assertEqual(server.api_url, server_config["api-url"])
        self.assertEqual(server.scope, server_config["scope"])
        self.assertEqual(server.api_token, server_config["token"])

    def test_server_configuration_default(self) -> None:
        """ConfigHandler.server_configuration() uses the default server."""
        configuration = self.valid_configuration()

        config_handler = self.build_config_handler(
            configuration, server_name=None
        )

        default_server = configuration[
            f"server:{configuration['General']['default-server']}"
        ]

        server = config_handler.server_configuration()
        self.assertEqual(server.api_url, default_server["api-url"])
        self.assertEqual(server.scope, default_server["scope"])
        self.assertEqual(server.api_token, default_server["token"])

    def test_server_configuration_kali(self) -> None:
        """ConfigHandler.server_configuration() return Kali."""
        config = self.valid_configuration()
        server_name = "kali"

        config_handler = self.build_config_handler(
            config, server_name=server_name
        )

        config_server = config[f"server:{server_name}"]

        server = config_handler.server_configuration()
        self.assertEqual(server.api_url, config_server["api-url"])
        self.assertEqual(server.scope, config_server["scope"])
        self.assertEqual(server.api_token, config_server["token"])

    def test_server_configuration_find_by_fqdn(self) -> None:
        """ConfigHandler.server_configuration() finds a section by FQDN."""
        config = self.valid_configuration()

        config_handler = self.build_config_handler(
            config, server_name="debusine.kali.org/kali"
        )
        config_server = config["server:kali"]
        server = config_handler.server_configuration()
        self.assertEqual(server.api_url, config_server["api-url"])
        self.assertEqual(server.scope, config_server["scope"])
        self.assertEqual(server.api_token, config_server["token"])

    def test_server_configuration_find_by_fqdn_wrong_scope(self) -> None:
        """ConfigHandler.server_configuration() requires FQDN/scope to match."""
        config = self.valid_configuration()
        config_handler = self.build_config_handler(
            config, server_name="debusine.kali.org/not-kali"
        )

        with self.assertRaisesRegex(
            ValueError,
            r"No Debusine client configuration for "
            r"'debusine\.kali\.org/not-kali'; "
            r"run 'debusine setup' to configure it",
        ):
            config_handler.server_configuration()

    def test_server_non_existing_error(self) -> None:
        """ConfigHandler._server_configuration('does-not-exist') aborts."""
        config = self.build_config_handler(self.valid_configuration())

        with self.assertRaisesRegex(
            ValueError,
            "No Debusine client configuration for 'does-not-exist'; "
            "run 'debusine setup' to configure it",
        ):
            config._server_configuration('does-not-exist')

    def test_server_incomplete_configuration_error(self) -> None:
        """ConfigHandler._server_configuration('incomplete-server') aborts."""
        config_parser = ConfigParser()
        config_parser.add_section('server:incomplete-server')

        config = self.build_config_handler(config_parser)

        with self.assertRaisesSystemExit(3):
            config._server_configuration('incomplete-server')

        self.assertEqual(
            self.stderr.getvalue(),
            'Missing required keys in the section '
            '[server:incomplete-server]: api-url, scope, token in '
            f'{config._config_file_path} .\n',
        )

    def test_server_extra_configuration_error(self) -> None:
        """ConfigHandler._server_configuration('too-many-keys') aborts."""
        config_parser = self.valid_configuration()
        server_name = "debian"
        server_section = f"server:{server_name}"

        config_parser[server_section]['port'] = '80'
        config_parser[server_section]['priority'] = 'high'

        config = self.build_config_handler(config_parser)

        with self.assertRaisesSystemExit(3):
            config._server_configuration(server_name)

        self.assertEqual(
            self.stderr.getvalue(),
            'Invalid keys in the section '
            f'[{server_section}]: port, priority in '
            f'{config._config_file_path} .\n',
        )

    def test_config_file_cannot_be_opened(self) -> None:
        """ConfigHandler.__init__() cannot open the file: aborts."""
        with self.assertRaisesSystemExit(3):
            ConfigHandler(
                config_file_path='/tmp/does_not_exist101001.ini',
                stderr=self.stderr,
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Utilities for assets."""

from pgpy import PGPKeyring
from pgpy.types import Armorable


# mypy complains that PGPKeyring and Armorable are untyped, which is true,
# but we can't fix that here.
class ExportablePGPKeyring(PGPKeyring, Armorable):  # type: ignore[misc]
    """
    An OpenPGP keyring that can be exported as a single public key block.

    Only public keys can be exported this way.
    """

    @property
    def magic(self) -> str:
        """The magic string identifier for this PGP type."""
        return "PUBLIC KEY BLOCK"

    def __bytearray__(self) -> bytearray:
        """Return the public key block as a byte array."""
        _bytes = bytearray()
        for fingerprint in self.fingerprints(
            keyhalf="public", keytype="primary"
        ):
            with self.key(fingerprint) as key:
                _bytes += key.__bytearray__()
        return _bytes

    def __bytes__(self) -> bytes:
        """Return the public key block as bytes."""
        return bytes(self.__bytearray__())



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Waiter utility: waits until success/failure and fails after timeout."""

import datetime as dt
import time as tm
from collections.abc import Callable
from functools import partialmethod
from typing import Any


class Waiter:
    """Waits for a function to return True or fails."""

    @staticmethod
    def _wait_for(
        expected: Any, timeout: float, func: Callable[[], Any]
    ) -> bool:
        """
        Execute func until it returns expected, or fails on timeout.

        :param expected: expected output of the func
        :param timeout: maximum number of seconds to wait
        :param func: function that gets executed
        :return: True if it succeeded, False if it timed out
        """
        start_at = dt.datetime.now(tz=dt.UTC)

        while func() != expected:
            if dt.datetime.now(tz=dt.UTC) - start_at > dt.timedelta(
                seconds=timeout
            ):
                return False

            tm.sleep(0.2)

        return True

    wait_for_success: partialmethod[bool] = partialmethod(_wait_for, True)
    wait_for_failure: partialmethod[bool] = partialmethod(_wait_for, False)



# Generated by Django 4.2.21 on 2025-09-03 13:05

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0015_alter_workspace_expiration_delay'),
    ]

    operations = [
        migrations.AddIndex(
            model_name='artifact',
            index=models.Index(
                models.F('workspace'),
                models.F('category'),
                include=('id',),
                name='db_artifact_workspace_category',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine views."""

import datetime as dt
from typing import Any, override

from django.contrib import messages
from django.http import (
    HttpRequest,
    HttpResponse,
    HttpResponseBase,
    HttpResponseRedirect,
)
from django.urls import reverse
from django.utils.safestring import SafeString
from django.views.generic import TemplateView, View

from debusine.db.context import context
from debusine.db.models import WorkRequest, Workspace
from debusine.web.views.base import BaseUIView
from debusine.web.views.table import NumberColumn, StringColumn, Table
from debusine.web.views.tables import WorkflowTable


class ScopeWorkspaceTable(Table[Workspace]):
    """Table showing scopes and workspaces."""

    scope = StringColumn("Scope")
    workspace = StringColumn("Workspace")
    role = StringColumn("Role")
    running = NumberColumn(
        SafeString(
            '<span title="Running workflows"'
            ' class="badge text-bg-info">R</span>'
        )
    )
    input_needed = NumberColumn(
        SafeString(
            '<span title="Input needed workflows"'
            ' class="badge text-bg-info">I</span>'
        )
    )
    completed = NumberColumn(
        SafeString(
            '<span title="Completed and aborted workflows"'
            ' class="badge text-bg-primary">C</span>'
        )
    )

    template_name = "web/_scope_workspace-list.html"


class HomepageView(BaseUIView, TemplateView):
    """Class for the homepage view."""

    base_template = "web/_base_debusine.html"
    template_name = "web/homepage.html"
    title = "Main page"

    @override
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """Return context_data with work_request_list and workspace_list."""
        ctx = super().get_context_data(**kwargs)

        if self.request.user.is_authenticated:
            workflows = (
                WorkRequest.objects.select_related(
                    "workspace__scope", "workflow_root"
                )
                .can_display(context.pc)
                .filter(created_by=self.request.user)
                .workflows()
                .filter(parent__isnull=True)
            )
            ctx["workflows_current"] = WorkflowTable(
                self.request,
                workflows.running(),
                prefix="current",
                preview=True,
            ).get_paginator(per_page=5)

            ctx["workflows_completed"] = WorkflowTable(
                self.request,
                workflows.filter(
                    status__in=(
                        WorkRequest.Statuses.COMPLETED,
                        WorkRequest.Statuses.ABORTED,
                    )
                ),
                prefix="completed",
                preview=True,
            ).get_paginator(per_page=5)

        ctx["workspaces"] = ScopeWorkspaceTable(
            self.request,
            Workspace.objects.can_display(context.pc)
            .select_related("scope")
            .order_by("scope", "name")
            .annotate_with_workflow_stats()
            .with_role_annotations(context.pc),
            prefix="workspaces",
        ).get_paginator(per_page=10)

        return ctx


class ToastView(BaseUIView, TemplateView):
    """Plumbing view to test toasts."""

    base_template = "web/_base_debusine.html"
    template_name = "web/toasts.html"
    title = "Test showing django messages"

    @override
    def dispatch(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        """Dispatch the request, setting appropriate response headers."""
        response = super().dispatch(request, *args, **kwargs)

        match request.GET.get("type"):
            case "debug":
                messages.debug(request, "This is a debugging message.")
            case "info":
                messages.info(request, "This is an informational message.")
            case "success":
                messages.success(request, "This message represents success.")
            case "warning":
                messages.warning(request, "This is supposed to be a warning.")
            case "error":
                messages.error(request, "An error has not occurred.")
            case None:
                pass
            case _:
                messages.error(request, "Unknown toast type.")

        return response


class ThemeSelectionView(View):
    """Select and save the current theme."""

    def post(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponse:
        """Set the selected theme."""
        value = request.POST.get("theme", "auto")
        next_url = request.POST.get("next", None)
        if next_url is None:
            next_url = reverse("homepage:homepage")
        response = HttpResponseRedirect(next_url)
        if value == "auto":
            response.delete_cookie("theme")
        else:
            response.set_cookie(
                "theme", value, httponly=False, max_age=dt.timedelta(days=3650)
            )
        return response



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to manage workers."""

import warnings
from typing import Any, NoReturn, override

from django.core.management import CommandParser

from debusine.server.management.commands import worker
from debusine.tasks.models import WorkerType


class Command(worker.Command):
    """Command to manage workers. E.g. enable and disable them."""

    help = 'Enables and disables workers'

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "--worker-type",
            choices=["external", "signing"],
            type=WorkerType,
            default=WorkerType.EXTERNAL,
        )
        parser.add_argument(
            'action',
            choices=['enable', 'disable'],
            help="Enable or disable the worker. Disabling a worker makes the "
            "worker's token disabled (worker cannot communicate with the "
            "server) and de-assigns work requests for the worker, which "
            "will be assigned to another worker. No attempt is made to stop "
            "any worker's operations but the server will not accept any "
            "results",
        )
        parser.add_argument('worker', help='Name of the worker to modify')

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Forward to `debusine-admin worker enable` or `... disable`."""
        action = options["action"]
        warnings.warn(
            f"The `debusine-admin manage_worker {action}` command has been "
            f"deprecated in favour of `debusine-admin worker {action}`",
            DeprecationWarning,
        )
        super().handle(*args, **options)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""SimpleSystemImageBuild Task, extending the SystemImageBuild ontology."""

import os
import shlex
import subprocess
from pathlib import Path
from typing import Any, override

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine import utils
from debusine.artifacts.local_artifact import DebianSystemImageArtifact
from debusine.artifacts.models import WorkRequestResults
from debusine.tasks.models import DiskImageFormat, SystemImageBuildData
from debusine.tasks.systembootstrap import SystemBootstrap


class SimpleSystemImageBuild(SystemBootstrap[SystemImageBuildData]):
    """Implement SimpleSystemImageBuild using debefivm-create."""

    _OUTPUT_SYSTEM_FILE = "system.img"
    _VAR_LIB_DPKG = "var_lib_dpkg"

    @override
    @classmethod
    def analyze_worker(cls) -> dict[str, Any]:
        """Report metadata for this task on this worker."""
        metadata = super().analyze_worker()

        available_key = cls.prefix_with_task_name("available")
        metadata[available_key] = utils.is_command_available("mmdebstrap")

        return metadata

    @override
    @classmethod
    def get_provided_worker_tags(cls) -> set[str]:
        tags = super().get_provided_worker_tags()
        if utils.is_command_available("mmdebstrap"):
            tags.add(wtags.CAP_PREFIX + "mmdebstrap")
        return tags

    @override
    def compute_scheduler_tags(
        self,
    ) -> tuple[ttags.TaskProvidedTags, wtags.TaskRequiredTags]:
        provided, required = super().compute_scheduler_tags()

        required.add(
            wtags.ProvenanceRequired.SYSTEM, {wtags.CAP_PREFIX + "mmdebstrap"}
        )

        return provided, required

    @override
    def can_run_on(self, worker_metadata: dict[str, Any]) -> bool:
        if not super().can_run_on(worker_metadata):
            return False

        available_key = self.prefix_with_task_name("available")
        if not worker_metadata.get(available_key, False):
            return False

        return True

    @override
    def _cmdline(self) -> list[str]:
        """
        Return debefivm-create command line.

        Use configuration of self.data.
        """
        chroot_source = self._chroot_sources_file
        cmd = [
            "debefivm-create",
            f"--architecture={self.data.bootstrap_options.architecture}",
            f"--release={self.data.bootstrap_repositories[0].suite}",
            f"--rootsize={self.data.disk_image.partitions[0].size}G",
            f"--output={self._OUTPUT_SYSTEM_FILE}",  # output file
            "--",
            f"{self.data.bootstrap_repositories[0].mirror}",
            "--verbose",
            # Set "find"'s cwd to "$1". Otherwise, find cwd is the
            # execute_directory as created by RunCommandTask._execute()
            # which is not readable by the mmdebstrap's subuid (find runs
            # under mmdebstrap's subuid already in --mode=unshare). In this
            # case find fails with
            # "Failed to restore initial working directory"
            (
                '--customize-hook=cd "$1" && '
                "find etc/apt/sources.list.d -type f -delete"
            ),
            (
                f"--customize-hook=upload {chroot_source} "
                "/etc/apt/sources.list.d/file.sources"
            ),
        ]

        # Begin deal with the keyrings
        # Upload keyrings needed in the chroot
        keyrings_dir = "/etc/apt/keyrings-debusine"
        cmd.append(f'--customize-hook=mkdir "$1{keyrings_dir}"')

        cmd += [
            f"--customize-hook=upload {keyring} {keyrings_dir}/{keyring.name}"
            for keyring in self._upload_keyrings
        ]

        # Add --keyring for each keyring downloaded by
        cmd += [f"--keyring={keyring}" for keyring in self._keyrings]

        cmd += [
            f"--include={package}"
            for repository in self.data.bootstrap_repositories
            if (package := repository.keyring_package)
        ]
        # End deal with the keyrings

        # customization_script
        if script := self._customization_script:
            # special case autopkgtest script
            # as it needs to run outside of the chroot
            if script == Path(
                "/usr/share/autopkgtest/setup-commands/setup-testbed"
            ):
                cmd.append(f"--customize-hook={script}")
            else:
                script_name = script.name
                cmd.extend(
                    [
                        f"--customize-hook=upload {script} /{script_name}",
                        f'--customize-hook=chmod 555 "$1/{script_name}"',
                        f'--customize-hook=chroot "$1" /{script_name}',
                        f'--customize-hook=rm "$1/{script_name}"',
                    ]
                )

        # Used by `upload_artifacts`
        cmd.extend(
            [
                (
                    "--customize-hook=download /etc/os-release "
                    f"{shlex.quote(self._OS_RELEASE_FILE)}"
                ),
                (
                    "--customize-hook=tar-out /var/lib/dpkg "
                    f"{shlex.quote(self._VAR_LIB_DPKG)}.tar"
                ),
            ]
        )

        overlay_dir = (
            "/usr/lib/python3/dist-packages/debusine/tasks/data/overlays"
        )
        for overlay, path, mode in (
            ("incus-agent", "lib/systemd/system/incus-agent.service", 0o644),
            ("incus-agent", "lib/systemd/incus-agent-setup", 0o755),
            ("incus-agent", "lib/udev/rules.d/99-incus-agent.rules", 0o644),
            (
                "systemd-boot",
                "etc/kernel/postinst.d/zz-update-systemd-boot",
                0o755,
            ),
            (
                "systemd-boot",
                "etc/kernel/postrm.d/zz-update-systemd-boot",
                0o755,
            ),
        ):
            cmd.extend(
                [
                    (
                        f"--customize-hook=upload "
                        f"{overlay_dir}/{overlay}/{path} /{path}"
                    ),
                    f'--customize-hook=chmod {mode:o} "$1/{path}"',
                ]
            )

        if variant := self.data.bootstrap_options.variant:
            cmd.append(f"--variant={variant}")

        extra_packages = self.data.bootstrap_options.extra_packages

        kernel = self.data.disk_image.kernel_package
        if kernel is None:
            # Only introduced in bullseye
            # We require the generic kernel for 9p support (#1027174)
            kernel = "linux-image-generic"
        extra_packages.append(kernel)

        assert extra_packages
        cmd.append("--include=" + ",".join(extra_packages))

        cmd.append(str(self._host_sources_file))

        return cmd

    @override
    def fetch_input(self, destination: Path) -> bool:  # noqa: U100
        """Do nothing: no artifacts need to be downloaded."""
        return True

    @override
    def upload_artifacts(
        self, execute_dir: Path, *, execution_result: WorkRequestResults
    ) -> None:
        """Upload generated artifacts."""
        if not self.debusine:
            raise AssertionError("self.debusine not set")

        if execution_result != WorkRequestResults.SUCCESS:
            return

        if self.data.disk_image.format == DiskImageFormat.QCOW2:
            system_file = execute_dir / f"{self.data.disk_image.filename}.qcow2"
            subprocess.check_call(
                [
                    "qemu-img",
                    "convert",
                    "-O",
                    "qcow2",
                    execute_dir / self._OUTPUT_SYSTEM_FILE,
                    system_file,
                ]
            )
        else:
            system_file = (
                execute_dir / f"{self.data.disk_image.filename}.tar.xz"
            )
            subprocess.check_call(
                [
                    "tar",
                    "--create",
                    "--auto-compress",
                    "--file",
                    system_file,
                    execute_dir / self._OUTPUT_SYSTEM_FILE,
                ]
            )

        bootstrap_options = self.data.bootstrap_options
        vendor = self._get_value_os_release(
            execute_dir / self._OS_RELEASE_FILE, "ID"
        )

        main_bootstrap_repository = self.data.bootstrap_repositories[0]
        codename = self._get_value_os_release(
            execute_dir / self._OS_RELEASE_FILE, "VERSION_CODENAME"
        )

        # /etc/os-release reports the testing release for unstable (#341)
        if main_bootstrap_repository.suite in ("unstable", "sid"):
            codename = "sid"

        components: list[str]
        if self.data.bootstrap_repositories[0].components:
            components = self.data.bootstrap_repositories[0].components
        elif self._chroot_sources_file:
            components = self._get_source_components(self._chroot_sources_file)
        else:
            raise AssertionError("self._chroot_sources_file not set")

        os.mkdir(execute_dir / self._VAR_LIB_DPKG)
        subprocess.check_call(
            [
                "tar",
                "-C",
                execute_dir / self._VAR_LIB_DPKG,
                "-xf",
                execute_dir / f"{self._VAR_LIB_DPKG}.tar",
            ],
        )
        pkglist = self._get_pkglist(execute_dir / self._VAR_LIB_DPKG)

        artifact = DebianSystemImageArtifact.create(
            system_file,
            data={
                "variant": bootstrap_options.variant,
                "architecture": bootstrap_options.architecture,
                "vendor": vendor,
                "codename": codename,
                "components": components,
                "pkglist": pkglist,
                # with_dev: with the current setup (in mmdebstrap),
                # the devices in /dev are always created
                "with_dev": True,
                "with_init": True,
                # TODO: Is "mirror" meant to be the first repository?
                # Or "mirrors" and list all of them?
                # Or we could duplicate all the bootstrap_repositories...
                "mirror": main_bootstrap_repository.mirror,
                "image_format": self.data.disk_image.format,
                "filesystem": self.data.disk_image.partitions[0].filesystem,
                "size": (self.data.disk_image.partitions[0].size * 10**9),
                "boot_mechanism": "efi",
            },
        )

        self.debusine.upload_artifact(
            artifact,
            workspace=self.workspace_name,
            work_request=self.work_request_id,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data migration to populate ``WorkRequest.workflow_root``."""

from django.db import migrations
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.migrations.state import StateApps
from django.db.models import Case, F, OuterRef, Q, Subquery, When

from debusine.artifacts.models import TaskTypes


def set_workflow_root(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Set ``WorkRequest.workflow_root``."""
    WorkRequest = apps.get_model("db", "WorkRequest")

    # This efficiently propagates workflow root information down from each
    # root.  The first update will update all roots, the second will update
    # all immediate children of roots, and so on.  It stops when there are
    # no more changes left to be made.
    while (
        WorkRequest.objects.filter(
            # Select work requests that are workflows or parts of workflows.
            (Q(task_type=TaskTypes.WORKFLOW) | Q(parent__isnull=False)),
            # Select work requests where workflow_root is not yet set.
            Q(workflow_root__isnull=True),
            # Select work requests that are workflow roots, or where we've
            # already set the workflow root for their parent.
            (Q(parent__isnull=True) | Q(parent__workflow_root__isnull=False)),
        ).update(
            workflow_root=Subquery(
                WorkRequest.objects.filter(id=OuterRef("id"))
                .annotate(
                    new_workflow_root=Case(
                        # If there is no parent, then this must be its own
                        # workflow root (because we selected only work
                        # requests that are workflows or have a parent).
                        When(parent__isnull=True, then=F("pk")),
                        # Otherwise, the parent's workflow root must already
                        # be set, and we can propagate that to its children.
                        default=F("parent__workflow_root"),
                    )
                )
                .values("new_workflow_root")[:1]
            )
        )
        > 0
    ):
        pass


def clear_workflow_root(
    apps: StateApps, schema_editor: BaseDatabaseSchemaEditor
) -> None:
    """Clear ``WorkRequest.workflow_root``."""
    WorkRequest = apps.get_model("db", "WorkRequest")
    WorkRequest.objects.update(workflow_root=None)


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0022_workrequest_workflow_root"),
    ]

    operations = [
        migrations.RunPython(set_workflow_root, clear_workflow_root),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""URLs for viewing artifacts."""

from django.urls import path

import debusine.web.views.artifacts as views

app_name = "artifacts"

urlpatterns = [
    path(
        "<int:artifact_id>/raw/<path:path>",
        views.FileDetailViewRaw.as_view(),
        name="file-raw",
    ),
    path(
        "<int:artifact_id>/file/<path:path>",
        views.FileDetailView.as_view(),
        name="file-detail",
    ),
    path(
        "<int:artifact_id>/download/<path:path>",
        views.DownloadPathView.as_view(),
        name="file-download",
    ),
    path(
        "<int:artifact_id>/download/",
        views.DownloadPathView.as_view(),
        name="download",
    ),
    path(
        "<int:artifact_id>/",
        views.ArtifactDetailView.as_view(),
        name="detail",
    ),
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command generate_service_key."""

import stat

from django.core.management import call_command
from django.test import TestCase as DjangoTestCase
from nacl.public import PrivateKey

from debusine.test import TestCase


class GenerateServiceKeyCommandTests(TestCase, DjangoTestCase):
    """Tests for the generate_service_key command."""

    def test_generate(self) -> None:
        """Write a new key with correct permissions."""
        temp_path = self.create_temporary_directory()
        key_path = temp_path / "0.key"

        with self.assertRaises(SystemExit) as raised:
            call_command("generate_service_key", str(key_path))
        self.assertEqual(raised.exception.code, 0)

        # The file has suitably-restrictive permissions.
        self.assertEqual(stat.S_IMODE(key_path.stat().st_mode), 0o600)
        # We can load the contents of the file as a private key.
        PrivateKey(key_path.read_bytes())



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""The qemu executor backend."""

import os
import platform
from pathlib import Path
from typing import NoReturn, override

from debusine import utils
from debusine.client.debusine import Debusine
from debusine.client.models import ArtifactResponse
from debusine.tasks.executors.base import (
    ExecutorImageCategory,
    ExecutorInterface,
    ExecutorStatistics,
    ImageImportError,
    ImageNotDownloadedError,
    InstanceInterface,
)
from debusine.tasks.executors.images import ImageCache
from debusine.tasks.models import BackendType


class QemuExecutor(ExecutorInterface, backend_name=BackendType.QEMU):
    """Support the qemu(1) executor."""

    system_image: ArtifactResponse
    working_directory: Path
    _image_cache: ImageCache
    _local_path: Path | None = None

    image_category = ExecutorImageCategory.IMAGE

    def __init__(
        self,
        debusine_api: Debusine,
        system_image_id: int,
        workspace: str,
        working_directory: Path,
    ) -> None:
        """
        Instantiate a QemuExecutor.

        :param debusine_api: The object to use the debusine client API.
        :param system_image_id: An artifact ID pointing to the system image.
        """
        self._image_cache = ImageCache(
            debusine_api, self.image_category, workspace=workspace
        )
        self.system_image = self._image_cache.image_artifact(system_image_id)
        self.working_directory = working_directory

    @override
    @classmethod
    def available(cls) -> bool:
        arch = platform.machine()
        if not utils.is_command_available(f"qemu-system-{arch}"):
            return False
        return os.access("/dev/kvm", os.W_OK)

    @override
    def download_image(self) -> str:
        if "python3-minimal" not in self.system_image.data["pkglist"]:
            raise ImageImportError(
                "Image doesn't contain a python3-minimal, required by qemu "
                "backend."
            )
        self._local_path = self._image_cache.download_image(self.system_image)
        return self.image_name()

    @override
    def image_name(self) -> str:
        """Return the path to the downloaded image."""
        if self._local_path is None:
            raise ImageNotDownloadedError("download_image() not yet called")
        return str(self._local_path)

    @override
    def create(self) -> NoReturn:
        """
        Create an QemuInstance using the downloaded image.

        Returns a new, stopped QemuInstance.
        """
        raise NotImplementedError("No Instance implemented for Qemu")

    @override
    def autopkgtest_virt_server(self) -> str:
        return "qemu"

    @override
    def autopkgtest_virt_args(self) -> list[str]:
        if self._local_path is None:
            raise ImageNotDownloadedError("download_image() not yet called")
        return [
            "--boot",
            self.system_image.data["boot_mechanism"],
            "--dpkg-architecture",
            self.system_image.data["architecture"],
            str(self._local_path),
        ]

    @override
    @classmethod
    def get_statistics(
        cls,
        instance: InstanceInterface | None,  # noqa: ARG003, U100
    ) -> ExecutorStatistics:
        """Return statistics for this executor."""
        # TODO: The autopkgtest-virt-server interface makes this very
        # difficult to implement.  See
        # https://salsa.debian.org/freexian-team/debusine/-/issues/671.
        return ExecutorStatistics()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Signon with external providers."""

# Code in this module is almost a standalone library, and may be split into one
# if there is a need of adopting it in multiple projects.
#
# Possible improvements that can be made to make this code more generic:
#
# * https://salsa.debian.org/freexian-team/debusine/-/issues/221
#   "Add a view to trigger linking other external auth providers to the current
#   user"
# * https://salsa.debian.org/freexian-team/debusine/-/issues/231
#   "Improvements to SSO-based account autocreation"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Functions to get system information."""

import subprocess
from functools import cache
from typing import Any

import psutil

import debusine.worker.tags as wtags
from debusine.tasks.models import WorkerType
from debusine.utils import debusine_version


def total_physical_memory() -> int:
    """Return bytes of RAM memory in system."""
    return psutil.virtual_memory().total


def cpu_count() -> int:
    """Return number of CPUs in the system."""
    if (count := psutil.cpu_count()) is None:
        return 1
    return count


def cpu_time() -> float:
    """Return total CPU time (user + system) used so far."""
    cpu_times = psutil.cpu_times()
    return cpu_times.user + cpu_times.system


@cache
def native_architecture() -> str | None:
    """Return the architecture of the system or None if dpkg not available."""
    try:
        return subprocess.check_output(
            ["dpkg", "--print-architecture"], text=True
        ).strip()
    except (FileNotFoundError, subprocess.CalledProcessError):
        return None


def kernel_supported_architectures() -> list[str] | None:
    """
    Return the architectures supported by the kernel (using arch-test).

    If arch-test fails: return None.
    """
    try:
        return sorted(
            subprocess.check_output(["arch-test", "-n"], text=True)
            .strip()
            .split("\n")
        )
    except subprocess.CalledProcessError:
        return None


def system_metadata(worker_type: WorkerType) -> dict[str, Any]:
    """Return system metadata formatted as a dictionary."""
    metadata = {
        "system:cpu_count": cpu_count(),
        "system:total_physical_memory": total_physical_memory(),
        "system:worker_type": worker_type,
        "system:worker_version": debusine_version(),
    }
    if native_arch := native_architecture():
        metadata.update(
            {
                "system:native_architecture": native_arch,
                "system:architectures": [native_arch],
            }
        )

    # More architectures might be supported by the kernel
    if (kernel_supported_archs := kernel_supported_architectures()) is not None:
        metadata["system:architectures"] = kernel_supported_archs

    return metadata


def system_provided_tags(worker_type: WorkerType) -> set[str]:
    """Return the tags provided by the worker system."""
    tags: set[str] = set()
    tags.add(wtags.WORKER_TYPE_PREFIX + worker_type)
    if native_arch := native_architecture():
        tags.add(wtags.BUILD_ARCH_PREFIX + native_arch)
    for arch in kernel_supported_architectures() or ():
        tags.add(wtags.BUILD_ARCH_PREFIX + arch)
    return tags



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the blhc task support on the worker."""

import itertools
from typing import override
from unittest.mock import call, patch

from debian.debian_support import Version

import debusine.tasks.tags as ttags
import debusine.worker.tags as wtags
from debusine.artifacts import BlhcArtifact
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianPackageBuildLog,
    WorkRequestResults,
)
from debusine.client.models import LookupSingleResponseArtifact
from debusine.tasks import Blhc, TaskConfigError
from debusine.tasks.models import BlhcDynamicData
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase
from debusine.test.test_utils import (
    create_remote_artifact,
    create_system_tarball_data,
)


class BlhcTests(ExternalTaskHelperMixin[Blhc], TestCase):
    """Test the Blhc class."""

    SAMPLE_TASK_DATA = {
        "input": {"artifact": 421},
        "build_architecture": "amd64",
        "environment": "debian/match:codename=bookworm",
    }

    build_log = DebianPackageBuildLog(
        source="hello",
        version="1.0",
        architecture="amd64",
        filename="hello_1.0_amd64.buildlog",
    )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.resolver.set_field_artifact(
            "environment",
            id=1,
            category=ArtifactCategory.SYSTEM_TARBALL,
            data=create_system_tarball_data(),
        )
        self.resolver.set_field_artifact(
            "input_artifact",
            id=421,
            category=ArtifactCategory.PACKAGE_BUILD_LOG,
            data=self.build_log,
        )

    def test_configure_fails_with_missing_required_data(
        self,
    ) -> None:
        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"input": {}})

    def test_configure_with_extra_flag(self) -> None:
        """Configuration included "extra_flags". Saved, no exceptions."""
        extra_flags = ["--bindnow"]

        directory = self.create_temporary_directory()
        (directory / "file.build").write_text("")

        self.configure_task(override={"extra_flags": extra_flags})

        self.patch_prepare_executor_instance()

        self.assertEqual(self.task.data.extra_flags, extra_flags)

        self.assertTrue(self.task.configure_for_execution(directory))

    def test_configure_with_invalid_flag(self) -> None:
        """Configuration included "extra_flags". Saved, no exceptions."""
        extra_flags = ["--unknown-flag"]

        directory = self.create_temporary_directory()
        (directory / "file.build").write_text("")

        with self.assertRaises(TaskConfigError):
            self.configure_task(override={"extra_flags": extra_flags})

    def test_configure_additional_properties_in_input(self) -> None:
        """Assert no additional properties in task_data input."""
        error_msg = "Extra inputs are not permitted"
        with self.assertRaisesRegex(TaskConfigError, error_msg):
            self.configure_task(
                task_data={
                    "input": {
                        "artifact": 5,
                        "some_additional_property": 87,
                    }
                }
            )

    def test_get_provided_worker_tags(self) -> None:
        self.assertCountEqual(
            self.task.get_provided_worker_tags(),
            [wtags.TASK_PREFIX + "worker:blhc:version:1"],
        )

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags(
            provide_system={ttags.SOURCE_PACKAGE_PREFIX + "hello"},
            require_system={
                wtags.BUILD_ARCH_PREFIX + "amd64",
                wtags.EXECUTOR_PREFIX + "unshare",
            },
        )

    def test_compute_dynamic_data(self) -> None:
        """Dynamic data receives relevant artifact IDs."""
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            BlhcDynamicData(
                environment_id=1,
                input_artifact_id=421,
                input_artifact_data=self.build_log,
                subject="hello",
                parameter_summary="hello_1.0 (amd64)",
                runtime_context=None,
            ),
        )

    def test_get_subject(self) -> None:
        self.assertEqual(self.get_subject(), "hello")

    def test_get_configuration_context(self) -> None:
        self.assertIsNone(self.get_configuration_context())

    def test_acceptable_input_categories(self) -> None:
        self.assertEqual(
            Blhc.inputs["input_artifact"].categories,  # type: ignore[attr-defined]
            (ArtifactCategory.PACKAGE_BUILD_LOG,),
        )

    def test_get_input_artifacts_ids(self) -> None:
        """Test get_input_artifacts_ids."""
        self.assertEqual(self.task.get_input_artifacts_ids(), [])

        self.task.dynamic_data = BlhcDynamicData(
            input_artifact_id=1, environment_id=2
        )
        self.assertEqual(self.task.get_input_artifacts_ids(), [1, 2])

    def test_configure_for_execution_from_artifact_error_no_build_logs(
        self,
    ) -> None:
        """configure_for_execution() no .build files: return False."""
        self.patch_prepare_executor_instance()
        download_directory = self.create_temporary_directory()
        (file1 := download_directory / "file1.dsc").write_text("")
        (file2 := download_directory / "file2.changes").write_text("")

        self.assertFalse(self.task.configure_for_execution(download_directory))

        files = sorted([str(file1), str(file2)])
        log_file_contents = (
            self.task.log_directory / "configure_for_execution.log"
        ).read_text()
        self.assertEqual(
            log_file_contents,
            f"No *.build files to be analyzed. Files: {files}\n",
        )

    def test_blhc_supports_debian_option_stretch(self) -> None:
        with patch.object(
            self.task,
            "get_package_version",
            return_value=Version("0.07+20161116+gitbf41976-1"),
        ):
            self.assertFalse(self.task._blhc_supports_debian_option())

    def test_blhc_supports_debian_option_buster(self) -> None:
        with patch.object(
            self.task, "get_package_version", return_value=Version("0.09-1")
        ):
            self.assertTrue(self.task._blhc_supports_debian_option())

    def test_cmdline_without_debian_option(self) -> None:
        self.task._blhc_target = self.create_temporary_file()

        with patch.object(
            self.task,
            "_blhc_supports_debian_option",
            autospec=True,
            return_value=False,
        ) as mock_blhc_supports_debian_option:
            cmdline = self.task._cmdline()

        self.assertEqual(cmdline, ["blhc", str(self.task._blhc_target)])
        mock_blhc_supports_debian_option.assert_called_once_with()

    def test_cmdline_with_debian_option(self) -> None:
        self.task._blhc_target = self.create_temporary_file()

        with patch.object(
            self.task,
            "_blhc_supports_debian_option",
            autospec=True,
            return_value=True,
        ) as mock_blhc_supports_debian_option:
            cmdline = self.task._cmdline()

        self.assertEqual(
            cmdline, ["blhc", "--debian", str(self.task._blhc_target)]
        )
        mock_blhc_supports_debian_option.assert_called_once_with()

    def test_cmdline_with_extra_flags(self) -> None:
        """Cmdline add extra_flags."""
        extra_flags = ["--bindnow"]
        self.configure_task(override={"extra_flags": extra_flags})
        self.task._blhc_target = self.create_temporary_file()

        with patch.object(
            self.task,
            "_blhc_supports_debian_option",
            autospec=True,
            return_value=True,
        ):
            cmdline = self.task._cmdline()

        self.assertIn(extra_flags[0], cmdline)
        self.assertEqual(cmdline[-1], str(self.task._blhc_target))

    def test_task_result_empty_file_success(self) -> None:
        """task_result() for an empty file return SUCCESS."""
        directory = self.create_temporary_directory()
        (directory / Blhc.CAPTURE_OUTPUT_FILENAME).write_text("")

        self.configure_task()
        self.assertEqual(
            self.task.task_result(0, directory), WorkRequestResults.SUCCESS
        )

    def test_task_result_failure_exit_code(self) -> None:
        """task_result() for an exit code of 8 return FAILURE."""
        directory = self.create_temporary_directory()
        (directory / Blhc.CAPTURE_OUTPUT_FILENAME).write_text("")

        self.configure_task()
        self.assertEqual(
            self.task.task_result(8, directory), WorkRequestResults.FAILURE
        )

    def test_upload_artifacts(self) -> None:
        """upload_artifact() and relation_create() is called."""
        exec_dir = self.create_temporary_directory()

        # Create file that will be attached when uploading the artifacts
        blhc_output = exec_dir / self.task.CAPTURE_OUTPUT_FILENAME
        blhc_output.write_text("NONVERBOSE BUILD:    Compiling gkrust\n")

        # Used to verify the relations
        self.task._source_artifacts_ids = [1]
        self.task.dynamic_data = BlhcDynamicData(
            input_artifact_id=2,
            input_artifact_data=self.build_log,
            environment_id=1,
        )

        # Debusine.upload_artifact is mocked to verify the call only
        debusine_mock = self.mock_debusine()

        workspace_name = "testing"

        uploaded_artifacts = [
            create_remote_artifact(id=10, workspace=workspace_name),
        ]

        debusine_mock.upload_artifact.side_effect = uploaded_artifacts

        # self.task.workspace_name is set by the Worker
        # and is the workspace that downloads the artifact
        # containing the files needed for Blhc
        self.task.workspace_name = workspace_name

        # The worker set self.task.work_request_id of the task
        work_request_id = 147
        self.task.work_request_id = work_request_id

        self.task.upload_artifacts(
            exec_dir, execution_result=WorkRequestResults.SUCCESS
        )

        # Debusine Mock upload_artifact expected calls
        upload_artifact_calls = []
        blhc_artifact = BlhcArtifact.create(
            blhc_output=blhc_output, build_log=self.build_log
        )

        upload_artifact_calls.append(
            call(
                blhc_artifact,
                workspace=workspace_name,
                work_request=work_request_id,
            )
        )

        # Debusine mock relation_create expected calls
        relation_create_calls = []
        for uploaded_artifact, source_artifact_id in itertools.product(
            uploaded_artifacts, self.task._source_artifacts_ids
        ):
            relation_create_calls.append(
                call(uploaded_artifact.id, source_artifact_id, "relates-to")
            )

        # Assert that the artifacts were uploaded and relations created
        debusine_mock.upload_artifact.assert_has_calls(upload_artifact_calls)
        debusine_mock.relation_create.assert_has_calls(relation_create_calls)

    def test_execute(self) -> None:
        """Test full (mocked) execution."""
        build_log = DebianPackageBuildLog(
            source="hello",
            version="1.0",
            architecture="i386",
            filename="hello_1.0_i386.buildlog",
        )
        self.configure_task(override={"input": {'artifact': 1}})
        self.task.work_request_id = 2
        self.task.workspace_name = "testing"
        self.task.dynamic_data = BlhcDynamicData(
            input_artifact_id=1, input_artifact_data=build_log, environment_id=2
        )
        download_directory = self.create_temporary_directory()

        debusine_mock = self.mock_debusine()
        debusine_mock.lookup_single.return_value = LookupSingleResponseArtifact(
            artifact=1
        )
        debusine_mock.download_artifact.return_value = True
        debusine_mock.upload_artifact.return_value = create_remote_artifact(
            id=2, workspace=self.task.workspace_name
        )

        self.assertTrue(self.task.fetch_input(download_directory))

        debusine_mock.download_artifact.assert_called_once_with(
            1, download_directory, tarball=False, workspace="testing"
        )

        f1_contents = "Compiling gkrust v0.1.0 (/build/rust)"
        (file1 := download_directory / "file.build").write_text(f1_contents)

        self.patch_prepare_executor_instance()

        self.assertTrue(self.task.configure_for_execution(download_directory))

        self.assertEqual(self.task._blhc_target, file1)
        with patch.object(
            self.task,
            "_blhc_supports_debian_option",
            autospec=True,
            return_value=True,
        ):
            self.assertEqual(
                self.task._cmdline(),
                [
                    "blhc",
                    "--debian",
                    str(file1),
                ],
            )

        # Create file that will be attached when uploading the artifacts
        blhc_output = download_directory / self.task.CAPTURE_OUTPUT_FILENAME
        blhc_output.write_text(
            "NONVERBOSE BUILD:    Compiling gkrust v0.1.0 (/build/rust)\n"
        )

        self.task.upload_artifacts(
            download_directory, execution_result=WorkRequestResults.SUCCESS
        )

        debusine_mock.upload_artifact.assert_called_once_with(
            BlhcArtifact.create(
                blhc_output=download_directory / Blhc.CAPTURE_OUTPUT_FILENAME,
                build_log=build_log,
            ),
            workspace=self.task.workspace_name,
            work_request=self.task.work_request_id,
        )
        debusine_mock.relation_create.assert_called_once_with(
            2, 1, "relates-to"
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the __main__ (Worker's CLI launcher)."""

import contextlib
import io
import signal
from pathlib import Path
from typing import override
from unittest import mock
from unittest.mock import AsyncMock

from debusine.test import TestCase
from debusine.worker.__main__ import main


class MainTests(TestCase):
    """Tests for __main__ (Worker's CLI launcher)."""

    @override
    def setUp(self) -> None:
        super().setUp()

        self.default_sigint_handler = signal.getsignal(signal.SIGINT)
        self.default_sigterm_handler = signal.getsignal(signal.SIGTERM)

        self.patch_config_handler()

    @override
    def tearDown(self) -> None:
        """Restore default signals (__main__.main() changes them)."""
        signal.signal(signal.SIGINT, self.default_sigint_handler)
        signal.signal(signal.SIGTERM, self.default_sigterm_handler)

        super().tearDown()

    def test_worker_version(self) -> None:
        stdout = io.StringIO()

        with (
            contextlib.redirect_stdout(stdout),
            self.assertRaisesSystemExit(0),
            mock.patch(
                "debusine.worker.__main__.debusine_version",
                return_value="0.1.0",
            ),
        ):
            main(["--version"])

        self.assertEqual("debusine-worker 0.1.0\n", stdout.getvalue())

    def test_worker_version_uninstalled(self) -> None:
        stdout = io.StringIO()

        with (
            contextlib.redirect_stdout(stdout),
            self.assertRaisesSystemExit(0),
            mock.patch(
                "debusine.worker.__main__.debusine_version", return_value=None
            ),
        ):
            main(["--version"])

        self.assertEqual(
            "debusine-worker uninstalled-version\n", stdout.getvalue()
        )

    def patch_config_handler(self) -> None:
        """Mock ConfigHandler to avoid trying to access config.ini."""
        config_handler_patcher = mock.patch(
            'debusine.worker._worker.ConfigHandler', autospec=True
        )
        self.config_handler_mocked = config_handler_patcher.start()
        instance = self.config_handler_mocked.return_value
        instance.log_level = None
        instance.working_directory = Path('/tmp')
        self.addCleanup(config_handler_patcher.stop)

    def patch_worker_main(self) -> mock.MagicMock:
        """Patch and return mock for Worker.main()."""
        patch = mock.patch(
            "debusine.worker.__main__.Worker.main", autospec=True
        )
        main_mocked = patch.start()
        self.addCleanup(patch.stop)

        return main_mocked

    def test_main_entry_point(self) -> None:
        """Worker's main entry point calls Worker().main()."""
        mocked_main = self.patch_worker_main()

        main([])

        mocked_main.assert_awaited_once()
        self.config_handler_mocked.assert_called()

    def patch_worker(self) -> mock.MagicMock:
        """Patch Worker class and returns its mock."""
        patch = mock.patch("debusine.worker.__main__.Worker")
        worker_mocked = patch.start()
        self.addCleanup(patch.stop)

        worker_mocked().main = AsyncMock()

        return worker_mocked

    def test_worker_parameters(self) -> None:
        """Worker main entry point calls run_worker() with the right params."""
        worker_mocked = self.patch_worker()

        main(
            [
                '--log-level',
                'DEBUG',
                '--working-directory',
                '/tmp',
            ]
        )

        worker_mocked.assert_called_with(
            log_level='DEBUG', working_directory='/tmp'
        )

    def test_default_worker_parameters(self) -> None:
        """Worker main entry point calls run_worker() with the right params."""
        worker_mocked = self.patch_worker()

        main([])

        worker_mocked.assert_called_with(log_level=None, working_directory=None)
